# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# January 22 2018, Christian E. Hopps <chopps@gmail.com>
#
# Copyright (c) 2018, Deutsche Telekom AG.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import getpass
import logging
from netconf import NSMAP, qmap
import netconf.util as ncutil
import netconf.error as ncerror
import netconf.server as ncserver

logger = logging.getLogger(__name__)
nc_server = None
NC_PORT = None
NC_DEBUG = True

mock_module = "urn:mock:module"


class MockMethods(object):
    NCFILTER = qmap("nc") + "filter"
    """This is an abstract class that is used to document the server methods functionality

    The server return not-implemented if the method is not found in the methods object,
    so feel free to use duck-typing here (i.e., no need to inherit)
    """

    def nc_append_capabilities(self, capabilities):  # pylint: disable=W0613
        """The server should append any capabilities it supports to capabilities"""
        ncutil.subelm(capabilities, "capability").text = mock_module
        ncutil.subelm(capabilities,
                      "capability").text = "urn:ietf:params:netconf:capability:xpath:1.0"

    def rpc_get(self, session, rpc, filter_or_none):  # pylint: disable=W0613
        data = ncutil.elm("data")
        cont = ncutil.subelm(data, "interfaces")
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "Ethernet0/0"))
        listval.append(ncutil.leaf_elm("shutdown", "true"))
        listval.append(ncutil.leaf_elm("state", "down"))
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "Ethernet0/1"))
        listval.append(ncutil.leaf_elm("shutdown", "false"))
        listval.append(ncutil.leaf_elm("state", "down"))
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "FastEthernet1/0"))
        listval.append(ncutil.leaf_elm("shutdown", "false"))
        listval.append(ncutil.leaf_elm("state", "up"))
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "FastEthernet1/1"))
        listval.append(ncutil.leaf_elm("shutdown", "false"))
        listval.append(ncutil.leaf_elm("state", "down"))

        return ncutil.filter_results(rpc, data, filter_or_none)

    def rpc_get_config(self, session, rpc, source_elm, filter_or_none):  # pylint: disable=W0613
        assert source_elm is not None
        if source_elm.find("nc:running", namespaces=NSMAP) is None:
            # Really this should be a different error its a bad value for source not missing
            raise ncerror.MissingElementProtoError(rpc, ncutil.qname("nc:running"))

        data = ncutil.elm("data")
        cont = ncutil.subelm(data, "interfaces")
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "Ethernet0/0"))
        listval.append(ncutil.leaf_elm("shutdown", "true"))
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "Ethernet0/1"))
        listval.append(ncutil.leaf_elm("shutdown", "false"))
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "FastEthernet1/0"))
        listval.append(ncutil.leaf_elm("shutdown", "false"))
        listval = ncutil.subelm(cont, "interface")
        listval.append(ncutil.leaf_elm("name", "FastEthernet1/1"))
        listval.append(ncutil.leaf_elm("shutdown", "false"))
        # Missing from operational
        listval.append(ncutil.leaf_elm("name", "GigabitEthernet2/0"))
        listval.append(ncutil.leaf_elm("shutdown", "false"))

        return ncutil.filter_results(rpc, data, filter_or_none)

    #---------------------------------------------------------------------------
    # These definitions will change to include required parameters like get and
    # get-config
    #---------------------------------------------------------------------------

    # XXX The API WILL CHANGE consider unfinished
    def rpc_copy_config(self, unused_session, rpc, *unused_params):
        raise ncerror.OperationNotSupportedProtoError(rpc)

    # XXX The API WILL CHANGE consider unfinished
    def rpc_delete_config(self, unused_session, rpc, *unused_params):
        raise ncerror.OperationNotSupportedProtoError(rpc)

    # XXX The API WILL CHANGE consider unfinished
    def rpc_edit_config(self, unused_session, rpc, *unused_params):
        raise ncerror.OperationNotSupportedProtoError(rpc)

    # XXX The API WILL CHANGE consider unfinished
    def rpc_lock(self, unused_session, rpc, *unused_params):
        raise ncerror.OperationNotSupportedProtoError(rpc)

    # XXX The API WILL CHANGE consider unfinished
    def rpc_unlock(self, unused_session, rpc, *unused_params):
        raise ncerror.OperationNotSupportedProtoError(rpc)


def init_mock_server():
    # logging.basicConfig(level=logging.DEBUG)

    if init_mock_server.server is not None:
        logger.error("XXX Called init_mock_server called multiple times")
    else:
        sctrl = ncserver.SSHUserPassController(username=getpass.getuser(), password="admin")
        init_mock_server.server = ncserver.NetconfSSHServer(
            server_ctl=sctrl,
            server_methods=MockMethods(),
            port=None,
            host_key="tests/host_key",
            debug=NC_DEBUG)
    return init_mock_server.server.port


init_mock_server.server = None

__author__ = 'Christian E. Hopps'
__date__ = 'January 22 2018'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# August 23 2017, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2017-2018, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes

import argparse
import logging
import os
import sys
from lxml import etree
from sshutil.server import from_private_key_file
from . import client


def parse_password_arg(password):
    if password:
        if password.startswith("env:"):
            unused, key = password.split(":", 1)
            password = os.environ[key]
        elif password.startswith("file:"):
            unused, path = password.split(":", 1)
            password = open(path).read().rstrip("\n")
    return password


def main(*margs):
    parser = argparse.ArgumentParser("Netconf Client Utility")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")
    parser.add_argument(
        '--edit-config',
        const="",
        nargs='?',
        help=("Perform <edit-config>. arg value is the method ('merge' or 'replace') " +
              " 'merge' if not specified. New config is in 'infile' or stdin"))
    parser.add_argument('--host', default="localhost", help='Netconf server hostname')
    parser.add_argument(
        '--get',
        const="",
        nargs='?',
        help="Perform <get>. arg value is xpath/xml filter or taken from infile if not specified")
    parser.add_argument(
        '--get-config',
        const="",
        nargs='?',
        help=
        "Perform <get-config>. arg value is xpath/xml filter or taken from infile if not specified")
    parser.add_argument(
        '--hello', action="store_true", help="Do hello and return capabilities of server.")
    parser.add_argument("-i", "--infile", help="File to read from")
    parser.add_argument(
        '-p',
        '--password',
        default=None,
        help='Password (or passphrase) use "env:" or "file:" prefix to specify password source')
    parser.add_argument(
        '-k', '--keyfile', default=None, help='SSH key use password to specify passphrase')
    # Deprecated now parse password args more functional
    parser.add_argument('--passenv', default=None, help=argparse.SUPPRESS)
    parser.add_argument('--port', type=int, default=830, help='Netconf server port')
    parser.add_argument("-q", "--quiet", action="store_true", help="Quiet operation")
    parser.add_argument('--source', default="running", help="Source for edit or get config")
    parser.add_argument('--timeout', type=float, help="Timeout for command in fractional seconds")
    parser.add_argument('-u', '--username', default="admin", help='Netconf username')
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose logging")
    args = parser.parse_args(*margs)

    if args.passenv and args.password:
        print("Only one of --password and --passenv allowed", file=sys.stderr)
        sys.exit(1)
    if args.passenv:
        args.password = os.environ[args.passenv]
    else:
        args.password = parse_password_arg(args.password)

    if args.keyfile:
        args.password = from_private_key_file(args.keyfile, password=args.password)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO)
    elif args.quiet:
        logging.basicConfig(level=logging.ERROR)
    else:
        logging.basicConfig(level=logging.WARNING)

    session = client.NetconfSSHSession(
        args.host, args.port, args.username, args.password, debug=args.debug)

    if args.hello:
        result = "\n".join(session.capabilities) + "\n"
    elif args.get is not None:
        result = session.get(args.get, args.timeout)
        result = "  " + etree.tounicode(result, pretty_print=True)
    elif args.get_config is not None:
        result = session.get_config(args.source, args.get_config, args.timeout)
        result = "  " + etree.tounicode(result, pretty_print=True)
    else:
        if args.infile:
            xml = open(args.infile).read()
        else:
            xml = sys.stdin.read()
        if not xml:
            print("Nothing to do.", file=sys.stderr)
            sys.exit(1)

        if args.edit_config is None:
            result = session.send_rpc(xml)[2]
        else:
            result = session.edit_config(args.source, args.edit_config, xml, args.timeout)
            result = etree.tounicode(result, pretty_print=True)
    sys.stdout.write(result)
    session.close()


if __name__ == "__main__":
    main()

__author__ = ''
__date__ = 'August 23 2017'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
# February 24 2018, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2018, Deutsche Telekom AG.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import argparse
import datetime
import logging
import os
import platform
import socket
import sys
import time
from netconf import error, server, util
from netconf import nsmap_add, NSMAP

nsmap_add("sys", "urn:ietf:params:xml:ns:yang:ietf-system")


def parse_password_arg(password):
    if password:
        if password.startswith("env:"):
            unused, key = password.split(":", 1)
            password = os.environ[key]
        elif password.startswith("file:"):
            unused, path = password.split(":", 1)
            password = open(path).read().rstrip("\n")
    return password


def date_time_string(dt):
    tz = dt.strftime("%z")
    s = dt.strftime("%Y-%m-%dT%H:%M:%S.%f")
    if tz:
        s += " {}:{}".format(tz[:-2], tz[-2:])
    return s


class SystemServer(object):
    def __init__(self, port, host_key, auth, debug):
        self.server = server.NetconfSSHServer(auth, self, port, host_key, debug)

    def close():
        self.server.close()

    def nc_append_capabilities(self, capabilities):  # pylint: disable=W0613
        """The server should append any capabilities it supports to capabilities"""
        util.subelm(capabilities,
                    "capability").text = "urn:ietf:params:netconf:capability:xpath:1.0"
        util.subelm(capabilities, "capability").text = NSMAP["sys"]

    def rpc_get(self, session, rpc, filter_or_none):  # pylint: disable=W0613
        """Passed the filter element or None if not present"""
        data = util.elm("nc:data")

        # if False: # If NMDA
        #     sysc = util.subelm(data, "system")
        #     sysc.append(util.leaf_elm("hostname", socket.gethostname()))

        #     # Clock
        #     clockc = util.subelm(sysc, "clock")
        #     tzname = time.tzname[time.localtime().tm_isdst]
        #     clockc.append(util.leaf_elm("timezone-utc-offset", int(time.timezone / 100)))

        sysc = util.subelm(data, "sys:system-state")
        platc = util.subelm(sysc, "sys:system")

        platc.append(util.leaf_elm("sys:os-name", platform.system()))
        platc.append(util.leaf_elm("sys:os-release", platform.release()))
        platc.append(util.leaf_elm("sys:os-version", platform.version()))
        platc.append(util.leaf_elm("sys:machine", platform.machine()))

        # Clock
        clockc = util.subelm(sysc, "sys:clock")
        now = datetime.datetime.now()
        clockc.append(util.leaf_elm("sys:current-datetime", date_time_string(now)))

        if os.path.exists("/proc/uptime"):
            with open('/proc/uptime', 'r') as f:
                uptime_seconds = float(f.readline().split()[0])
            boottime = time.time() - uptime_seconds
            boottime = datetime.datetime.fromtimestamp(boottime)
            clockc.append(util.leaf_elm("sys:boot-datetime", date_time_string(boottime)))

        return util.filter_results(rpc, data, filter_or_none, self.server.debug)

    def rpc_get_config(self, session, rpc, source_elm, filter_or_none):  # pylint: disable=W0613
        """Passed the source element"""
        data = util.elm("nc:data")
        sysc = util.subelm(data, "sys:system")
        sysc.append(util.leaf_elm("sys:hostname", socket.gethostname()))

        # Clock
        clockc = util.subelm(sysc, "sys:clock")
        # tzname = time.tzname[time.localtime().tm_isdst]
        clockc.append(util.leaf_elm("sys:timezone-utc-offset", int(time.timezone / 100)))

        return util.filter_results(rpc, data, filter_or_none)

    def rpc_system_restart(self, session, rpc, *params):
        raise error.AccessDeniedAppError(rpc)

    def rpc_system_shutdown(self, session, rpc, *params):
        raise error.AccessDeniedAppError(rpc)


def main(*margs):

    parser = argparse.ArgumentParser("Example System Server")
    parser.add_argument("--debug", action="store_true", help="Enable debug logging")
    parser.add_argument(
        "--password", default="admin", help='Use "env:" or "file:" prefix to specify source')
    parser.add_argument('--port', type=int, default=8300, help='Netconf server port')
    parser.add_argument("--username", default="admin", help='Netconf username')
    args = parser.parse_args(*margs)

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    args.password = parse_password_arg(args.password)
    host_key = os.path.dirname(__file__) + "/server-key"

    auth = server.SSHUserPassController(username=args.username, password=args.password)
    s = SystemServer(args.port, host_key, auth, args.debug)

    if sys.stdout.isatty():
        print("^C to quit server")
    try:
        while True:
            time.sleep(1)
    except Exception:
        print("quitting server")

    s.close()


if __name__ == "__main__":
    main()

__author__ = 'Christian Hopps'
__date__ = 'February 24 2018'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# March 31 2015, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import copy
import logging
from lxml import etree
from netconf import NSMAP
from netconf import error

# Tries to somewhat implement RFC6241 filtering
logger = logging.getLogger(__name__)


def qname(tag):
    try:
        return etree.QName(tag)
    except ValueError:
        prefix, base = tag.split(":")
        return etree.QName(NSMAP[prefix], base)


def elm(tag, attrib=None, **extra):
    if attrib is None:
        attrib = dict()
    return etree.Element(qname(tag), attrib, **extra)


def leaf_elm(tag, value, attrib=None, **extra):
    e = elm(tag, attrib, **extra)
    e.text = str(value)
    return e


# Create another name for leaf_elm function
leaf = leaf_elm


def subelm(pelm, tag, attrib=None, **extra):
    if attrib is None:
        attrib = dict()
    return etree.SubElement(pelm, qname(tag), attrib, **extra)


def is_selection_node(felm):
    ftext = felm.text
    return ftext is None or not ftext.strip()


def xpath_filter_result(data, xpath):
    """Filter a result given an xpath expression.

    :param data: The nc:data result element.
    :param xpath: The xpath expression string.
    :returns: New nc:data result element pruned by the xpath expression.

    >>> xml = '''
    ... <data>
    ...   <devs>
    ...     <dev>
    ...       <name>dev1</name>
    ...       <slots>1</slots>
    ...     </dev>
    ...     <dev>
    ...       <name>dev2</name>
    ...       <slots>2</slots>
    ...     </dev>
    ...     <dev>
    ...       <name>dev3</name>
    ...       <slots>3</slots>
    ...     </dev>
    ...   </devs>
    ... </data>
    ... '''
    >>> data = etree.fromstring(xml.replace(' ', '').replace('\\n', ''))
    >>> result = xpath_filter_result(data, "/devs/dev")
    >>> etree.tounicode(result)
    '<data><devs><dev><name>dev1</name><slots>1</slots></dev><dev><name>dev2</name><slots>2</slots></dev><dev><name>dev3</name><slots>3</slots></dev></devs></data>'
    >>> result = xpath_filter_result(data, "/devs/dev[name='dev1']")
    >>> etree.tounicode(result)
    '<data><devs><dev><name>dev1</name><slots>1</slots></dev></devs></data>'
    >>> result = xpath_filter_result(data, "/devs/dev[name='dev2']")
    >>> etree.tounicode(result)
    '<data><devs><dev><name>dev2</name><slots>2</slots></dev></devs></data>'
    >>> result = xpath_filter_result(data, "/devs/dev[name='dev2'] | /devs/dev[name='dev1']")
    >>> etree.tounicode(result)
    '<data><devs><dev><name>dev1</name><slots>1</slots></dev><dev><name>dev2</name><slots>2</slots></dev></devs></data>'
    >>> result = xpath_filter_result(data, "/devs/dev[name='dev1'] | /devs/dev[name='dev2']")
    >>> etree.tounicode(result)
    '<data><devs><dev><name>dev1</name><slots>1</slots></dev><dev><name>dev2</name><slots>2</slots></dev></devs></data>'
    >>> result = xpath_filter_result(data, "/devs/dev[name='dev1'] | /devs/dev[slots='2']")
    >>> etree.tounicode(result)
    '<data><devs><dev><name>dev1</name><slots>1</slots></dev><dev><name>dev2</name><slots>2</slots></dev></devs></data>'
    """

    # First get a copy we can safely modify.
    data = copy.deepcopy(data)

    results = []
    children = []

    # XXX Need to reset the namespace declarations to those found in the context of the filter node.

    # Have to re-root the children to avoid having to match "/nc:data"
    for child in data.getchildren():
        data.remove(child)
        children.append(child)
        newtree = etree.ElementTree(child)
        results.extend(newtree.xpath(xpath, namespaces=NSMAP))

    # Add the children of data back.
    for child in children:
        data.append(child)

    # Mark the tree up
    for result in results:
        # Mark all children
        for e in result.iterdescendants():
            e.attrib['__filter_marked__'] = ""
        # Mark this element and all parents
        while result is not data:
            result.attrib['__filter_marked__'] = ""
            result = result.getparent()

    def prunedecendants(e):
        for child in e.getchildren():
            if '__filter_marked__' not in child.attrib:
                e.remove(child)
            else:
                prunedecendants(child)
                del child.attrib['__filter_marked__']

    prunedecendants(data)

    return data


def filter_results(rpc, data, filter_or_none, debug=False):
    """Check for a user filter and prune the result data accordingly.

    :param rpc: An RPC message element.
    :param data: The data to filter.
    :param filter_or_none: Filter element or None.
    :type filter_or_none: `lxml.Element`
    """
    if filter_or_none is None:
        return data

    if 'type' not in filter_or_none.attrib or filter_or_none.attrib['type'] == "subtree":
        # Check for the pathalogical case of empty filter since that's easy to implement.
        if not filter_or_none.getchildren():
            return elm("data")
        # xpf = Convert subtree filter to xpath!
        logger.warning("Filtering with subtree not implemented yet.")
        raise error.OperationNotSupportedProtoError(rpc)
    elif filter_or_none.attrib['type'] == "xpath":
        if 'select' not in filter_or_none.attrib:
            raise error.MissingAttributeProtoError(rpc, filter_or_none, "select")
        xpf = filter_or_none.attrib['select']
    else:
        msg = "unexpected type: " + str(filter_or_none.attrib['type'])
        raise error.BadAttributeProtoError(rpc, filter_or_none, "type", message=msg)

    logger.debug("Filtering on xpath expression: %s", str(xpf))
    return xpath_filter_result(data, xpf)


def filter_tag_match(filter_tag, elm_tag):
    fqname = etree.QName(filter_tag)
    eqname = qname(elm_tag)
    if not fqname.namespace:
        return fqname.localname == eqname.localname
    return fqname == eqname


def filter_node_match_no_value(filter_node, match_elm):
    # First check to see if tag matches.
    if not filter_tag_match(filter_node.tag, match_elm.tag):
        return False

    # Next check for attribute matches.
    # XXX does this need to filter out namespace attributes?
    if filter_node.attrib and filter_node.attrib != match_elm.attrib:
        return False

    return True


def filter_node_match(filter_node, match_elm):
    """Given a filter node element and a nodename and attribute dictionary
    return true if the filter element matches the elmname, attributes and value
    (if not None).

    The filter element can use a wildcard namespace or a specific namespace
    the attributes can be missing from the filter node but otherwise must match
    and the value is only checked for a match if it is not None.
    """
    if not filter_node_match_no_value(filter_node, match_elm):
        return False

    # Finally check for matching value.
    ftext = filter_node.text
    if ftext is None:
        return True

    ftext = ftext.strip()
    if not ftext:
        return True

    return ftext == match_elm.text


def filter_leaf_values(fcontain_elm, dest_node, leaf_elms, append_to):
    """Given a containment element (or None) verify that all leaf elements
    in leaf_elms either match, have corresponding selection nodes (empty)
    or are not present.

    Additionally the correct leaf data will be added to dest_node, and dest_node
    will be appended to append_to if append_to is not None.

    The return value with be True, False, or a possibly empty set of selection/containment nodes
    The only failing value is False, if True is returned then the caller should include all
    containment sibling nodes, otherwise the caller should process the list of containment/selection
    nodes.
    """
    children = fcontain_elm.getchildren() if fcontain_elm is not None else []
    selected_elms = []
    if not children:
        selected_elms = leaf_elms

    # Now look at all the leaf filter selector or match nodes
    include_all_leaves = True
    othernodes = []
    for felm in children:
        fchildren = felm.getchildren()
        for lelm in leaf_elms:
            if fchildren:
                # Verify that this doesn't match a leaf node.
                if filter_node_match_no_value(felm, lelm):
                    # XXX this is an error we should raise some exception.
                    return False
                continue
            elif filter_node_match(felm, lelm):
                if not felm.text:
                    # This was a selection node.
                    include_all_leaves = False

                selected_elms.append(lelm)
                break
        else:
            if fchildren:
                # This is OK we verified a containment filter didn't match leaf by getting here.
                if felm.text:
                    # XXX verify that there is no text on this node, report violation?
                    return False

                # Track selection/filter nodes
                include_all_leaves = False
                othernodes.append(felm)
            elif not felm.text:
                # This is OK as it means this is a selection node include it in othernodes
                include_all_leaves = False
                othernodes.append(felm)
            else:
                # We've exhausted all leaf elements to match this leaf filter so we failed.
                return False

    # Everything matched so add in the leaf data.
    if append_to is not None:
        append_to.append(dest_node)

    if include_all_leaves:
        dest_node.extend(leaf_elms)
    else:
        dest_node.extend(selected_elms)

    if include_all_leaves:
        return True
    return othernodes


def filter_containment_iter(fcontain_elm, dest_node, containment_nodes, leaf_elms, append_to):
    """Given a containment filter node (or None) verify that all leaf elements
    either match, have corresponding selection nodes (empty) or are not present.

    If all leaf criteria are met then the iterator will return a triple of
    (new_filter_node, new_dest_node, new_data). new_filter_node corresponds to the
    matched containment node which is returned in new_dest_node, and new_data will be
    an element corresponding to the passed in dest_node.

    These should be processed by calling filter_containment_iter again.

    Additionally the correct leaf data will be added to dest_node, and dest_node
    will be appended to append_to if append_to is not None.

    This implements RFC6241 section 6.2.5
    """
    # No containment node so add everything.
    if fcontain_elm is None:
        # Add in the leaf data
        for e in leaf_elms:
            dest_node.append(e)

        # Append the match_node to the data
        if append_to is not None:
            append_to.append(dest_node)

        for node in containment_nodes:
            yield None, copy.copy(node), dest_node

    else:
        othernodes = filter_leaf_values(fcontain_elm, dest_node, leaf_elms, append_to)
        if othernodes is False:
            # No match
            pass
        elif othernodes is True:
            # All leaf values have matched and have been added and we should include all containers
            for node in containment_nodes:
                yield None, copy.copy(node), dest_node
        else:
            for felm in othernodes:
                for node in containment_nodes:
                    if filter_node_match_no_value(felm, node):
                        yield felm, copy.copy(node), dest_node


def filter_leaf_allows_add(filter_elm, tag, data, value):
    if filter_leaf_allows(filter_elm, tag, value):
        data.append(leaf_elm(tag, value))
        return True
    return False


def filter_leaf_allows(filter_elm, xpath, value):
    """Check the value at the xpath specified leaf matches the value.

    - If filter_elm is None then allow.
    - If there is no xpath element then allow if there are no other children.
    - XXX what about xpath that has embedded predicates!
      perhaps what we want to call this is a normal path not an xpath.
    """
    if filter_elm is None:
        return True

    # If there are no children then allow everything.
    if not filter_elm.getchildren():
        return True

    # No match or multiple matches not allowed for leaf.
    flist = filter_elm.xpath(xpath, namespaces=NSMAP)
    if not flist or len(flist) > 1:
        return False
    felm = flist[0]

    # No children for leaf allowed (leaf)
    if felm.getchildren():
        return False

    # Allowed if empty or if value matches.
    if not felm.text or felm.text == str(value):
        return True

    return False


def filter_list_iter(filter_list, key_xpath, keys):
    """Return key, elm pairs that are allowed by keys using the values found using
    the given key_xpath"""
    # If we have no filter elm then return all keys.
    if filter_list is None:
        for key in keys:
            yield key, None

    try:
        # If this an element then make it a list of elements
        filter_list.xpath  # pylint: disable=W0104
        filter_list = [filter_list]
    except AttributeError:
        pass

    for filter_elm in filter_list:
        filter_elms = [x for x in filter_elm.xpath(key_xpath, namespaces=NSMAP)]
        filter_keys = [x.text for x in filter_elms]
        if not filter_keys:
            for key in keys:
                yield key, filter_elm
        else:
            # Now walk our keys returning any that are in the filter list.
            for key in keys:
                if key in filter_keys:
                    yield key, filter_elm
                # try:
                #     idx = filter_keys.index(str(key))
                #     yield key, filter_elm
                # except ValueError:
                #     pass


__author__ = 'Christian Hopps'
__date__ = 'March 31 2015'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# February 19 2015, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import io
import logging
import os
import sys
import threading
import paramiko as ssh
from lxml import etree
import sshutil.server

from netconf import base
import netconf.error as ncerror
from netconf import NSMAP
from netconf import qmap
from netconf import util

if sys.platform == 'win32' and sys.version_info < (3, 5):
    import backports.socketpair  # pylint: disable=E0401,W0611

logger = logging.getLogger(__name__)

try:
    import pam
    have_pam = True
except ImportError:
    have_pam = False


class SSHAuthorizedKeysController(ssh.ServerInterface):
    """An implementation of paramiko `ServerInterface` that utilizes users
    authorized keys file for authentication.

    :param users: A list of usernames whose authorized keys will allow access.
    """

    def __init__(self, users=None):
        self.event = threading.Event()
        self.users = users
        self.users_keys = {}
        if have_pam:
            self.pam = pam.pam()
        else:
            self.pam = None

    def get_user_auth_keys(self, username):
        """Parse the users's authorized_keys file if any to look for authorized keys"""
        if username in self.users_keys:
            return self.users_keys[username]

        self.users_keys[username] = []

        userdir = os.path.expanduser("~" + username)
        if not userdir:
            return self.users_keys[username]

        keyfile = os.path.join(userdir, ".ssh/authorized_keys")
        if not keyfile or not os.path.exists(keyfile):
            return self.users_keys[username]

        with open(keyfile) as f:
            for line in f.readlines():
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                values = [x.strip() for x in line.split()]

                exp = None
                try:
                    int(values[0])  # bits value?
                except ValueError:
                    # Type 1 or type 2, type 1 is bits in second value
                    options_ktype = values[0]
                    try:
                        int(values[1])  # bits value?
                    except ValueError:
                        # type 2 with options
                        ktype = options_ktype
                        data = values[1]
                    else:
                        # Type 1 no options.
                        exp = int(values[1])
                        data = values[2]
                else:
                    # Type 1 no options.
                    exp = int(values[1])
                    data = values[2]

                # XXX For now skip type 1 keys
                if exp is not None:
                    continue

                if data:
                    import base64
                    if ktype == "ssh-rsa":
                        key = ssh.RSAKey(data=base64.decodebytes(data.encode('ascii')))
                    elif ktype == "ssh-dss":
                        key = ssh.DSSKey(data=base64.decodebytes(data.encode('ascii')))
                    else:
                        key = None
                    if key:
                        self.users_keys[username].append(key)
        return self.users_keys[username]

    def get_allowed_auths(self, username):
        # This is only called after the user fails some other authentication type.
        if self.users is None:
            users = [username]
        else:
            users = self.users
        allowed = []
        if username in users:
            if self.pam:
                allowed.append("password")

            if self.get_user_auth_keys(username):
                allowed.append("publickey")

        allowed = ",".join(allowed)
        logger.debug("Allowed methods for user %s: %s", str(username), allowed)
        return allowed

    def check_auth_none(self, unused_username):
        return ssh.AUTH_FAILED

    def check_auth_publickey(self, username, key):
        if not self.get_user_auth_keys(username):
            return ssh.AUTH_FAILED
        for ukey in self.users_keys[username]:
            if ukey == key:
                return ssh.AUTH_SUCCESSFUL
        return ssh.AUTH_FAILED

    def check_auth_password(self, username, password):
        # Don't allow empty user or empty passwords
        if not username or not password:
            return ssh.AUTH_FAILED
        if self.pam and self.pam.authenticate(username, password):
            return ssh.AUTH_SUCCESSFUL
        return ssh.AUTH_FAILED

    def check_channel_request(self, kind, chanid):
        if kind == "session":
            return ssh.OPEN_SUCCEEDED
        return ssh.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_channel_subsystem_request(self, channel, name):
        self.event.set()
        return name == "netconf"


# Backward compat
SSHAuthController = SSHAuthorizedKeysController


class SSHUserPassController(ssh.ServerInterface):
    """An implementation of paramiko `ServerInterface` that authorizes a single user
    and password.

    :param username: The username to allow.
    :param password: The password to allow.
    """

    def __init__(self, username=None, password=None):
        self.username = username
        self.password = password
        self.event = threading.Event()

    def get_allowed_auths(self, username):
        del username  # unused
        return "password"

    def check_auth_none(self, username):
        del username  # unused
        return ssh.AUTH_FAILED

    def check_auth_password(self, username, password):
        if self.username == username and self.password == password:
            return ssh.AUTH_SUCCESSFUL
        return ssh.AUTH_FAILED

    def check_channel_request(self, kind, chanid):
        if kind == "session":
            return ssh.OPEN_SUCCEEDED
        return ssh.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_channel_subsystem_request(self, channel, name):
        self.event.set()
        return name == "netconf"


class NetconfServerSession(base.NetconfSession):
    """Netconf Server-side session with a client.

    This object will be passed to a the server RPC methods.
    """
    handled_rpc_methods = set(["close-session", "kill-session"])

    def __init__(self, channel, server, unused_extra_args, debug):
        self.server = server

        sid = server._allocate_session_id()
        if debug:
            logger.debug("NetconfServerSession: Creating session-id %s", str(sid))

        self.methods = server.server_methods
        super(NetconfServerSession, self).__init__(channel, debug, sid)
        super(NetconfServerSession, self)._open_session(True)

        if self.debug:
            logger.debug("%s: Client session-id %s created", str(self), str(sid))

    def __del__(self):
        self.close()
        super(NetconfServerSession, self).__del__()

    def __str__(self):
        return "NetconfServerSession(sid:{})".format(self.session_id)

    def close(self):
        """Close the servers side of the session."""
        # XXX should be invoking a method in self.methods?
        if self.debug:
            logger.debug("%s: Closing.", str(self))

        try:
            super(NetconfServerSession, self).close()
        except EOFError:
            if self.debug:
                logger.debug("%s: EOF error while closing", str(self))

        if self.debug:
            logger.debug("%s: Closed.", str(self))

    # ----------------
    # Internal Methods
    # ----------------

    def _send_rpc_reply(self, rpc_reply, origmsg):
        """Send an rpc-reply to the client. This is should normally not be called
        externally the return value from the rpc_* methods will be returned
        using this method.
        """
        reply = etree.Element(qmap('nc') + "rpc-reply", attrib=origmsg.attrib, nsmap=origmsg.nsmap)
        try:
            rpc_reply.getchildren  # pylint: disable=W0104
            reply.append(rpc_reply)
        except AttributeError:
            reply.extend(rpc_reply)
        ucode = etree.tounicode(reply, pretty_print=True)
        if self.debug:
            logger.debug("%s: Sending RPC-Reply: %s", str(self), str(ucode))
        self.send_message(ucode)

    def _rpc_not_implemented(self, unused_session, rpc, *unused_params):
        if self.debug:
            msg_id = rpc.get('message-id')
            logger.debug("%s: Not Impl msg-id: %s", str(self), msg_id)
        raise ncerror.OperationNotSupportedProtoError(rpc)

    def _send_rpc_reply_error(self, error):
        self.send_message(error.get_reply_msg())

    def _reader_exits(self):
        if self.debug:
            logger.debug("%s: Reader thread exited.", str(self))
        return

    def _reader_handle_message(self, msg):
        if not self.session_open:
            return

        # Any error with XML encoding here is going to cause a session close
        # Technically we should be able to return malformed message I think.
        try:
            tree = etree.parse(io.BytesIO(msg.encode('utf-8')))
            if not tree:
                raise ncerror.SessionError(msg, "Invalid XML from client.")
        except etree.XMLSyntaxError:
            logger.warning("Closing session due to malformed message")
            raise ncerror.SessionError(msg, "Invalid XML from client.")

        rpcs = tree.xpath("/nc:rpc", namespaces=NSMAP)
        if not rpcs:
            raise ncerror.SessionError(msg, "No rpc found")

        for rpc in rpcs:
            try:
                msg_id = rpc.get('message-id')
                if self.debug:
                    logger.debug("%s: Received rpc message-id: %s", str(self), msg_id)
            except (TypeError, ValueError):
                raise ncerror.SessionError(msg, "No valid message-id attribute found")

            try:
                # Get the first child of rpc as the method name
                rpc_method = rpc.getchildren()
                if len(rpc_method) != 1:
                    if self.debug:
                        logger.debug("%s: Bad Msg: msg-id: %s", str(self), msg_id)
                    raise ncerror.MalformedMessageRPCError(rpc)
                rpc_method = rpc_method[0]

                rpcname = rpc_method.tag.replace(qmap('nc'), "")
                params = rpc_method.getchildren()
                paramslen = len(params)

                if self.debug:
                    logger.debug("%s: RPC: %s: paramslen: %s", str(self), rpcname, str(paramslen))

                if rpcname == "close-session":
                    # XXX should be RPC-unlocking if need be
                    if self.debug:
                        logger.debug("%s: Received close-session msg-id: %s", str(self), msg_id)
                    self._send_rpc_reply(etree.Element("ok"), rpc)
                    self.close()
                    # XXX should we also call the user method if it exists?
                    return
                elif rpcname == "kill-session":
                    # XXX we are supposed to cleanly abort anything underway
                    if self.debug:
                        logger.debug("%s: Received kill-session msg-id: %s", str(self), msg_id)
                    self._send_rpc_reply(etree.Element("ok"), rpc)
                    self.close()
                    # XXX should we also call the user method if it exists?
                    return
                elif rpcname == "get":
                    # Validate GET parameters

                    if paramslen > 1:
                        # XXX need to specify all elements not known
                        raise ncerror.MalformedMessageRPCError(rpc)
                    if params and not util.filter_tag_match(params[0], "nc:filter"):
                        raise ncerror.UnknownElementProtoError(rpc, params[0])
                    if not params:
                        params = [None]
                elif rpcname == "get-config":
                    # Validate GET-CONFIG parameters

                    if paramslen > 2:
                        # XXX Should be ncerror.UnknownElementProtoError? for each?
                        raise ncerror.MalformedMessageRPCError(rpc)
                    source_param = rpc_method.find("nc:source", namespaces=NSMAP)
                    if source_param is None:
                        raise ncerror.MissingElementProtoError(rpc, util.qname("nc:source"))
                    filter_param = None
                    if paramslen == 2:
                        filter_param = rpc_method.find("nc:filter", namespaces=NSMAP)
                        if filter_param is None:
                            unknown_elm = params[0] if params[0] != source_param else params[1]
                            raise ncerror.UnknownElementProtoError(rpc, unknown_elm)
                    params = [source_param, filter_param]

                #------------------
                # Call the method.
                #------------------

                try:
                    # Handle any namespaces or prefixes in the tag, other than
                    # "nc" which was removed above. Of course, this does not handle
                    # namespace collisions, but that seems reasonable for now.
                    rpcname = rpcname.rpartition("}")[-1]
                    method_name = "rpc_" + rpcname.replace('-', '_')
                    method = getattr(self.methods, method_name, self._rpc_not_implemented)
                    if self.debug:
                        logger.debug("%s: Calling method: %s", str(self), method_name)
                    reply = method(self, rpc, *params)
                    self._send_rpc_reply(reply, rpc)
                except NotImplementedError:
                    raise ncerror.OperationNotSupportedProtoError(rpc)
            except ncerror.MalformedMessageRPCError as msgerr:
                if self.new_framing:
                    if self.debug:
                        logger.debug("%s: MalformedMessageRPCError: %s", str(self), str(msgerr))
                    self.send_message(msgerr.get_reply_msg())
                else:
                    # If we are 1.0 we have to simply close the connection
                    # as we are not allowed to send this error
                    logger.warning("Closing 1.0 session due to malformed message")
                    raise ncerror.SessionError(msg, "Malformed message")
            except ncerror.RPCServerError as error:
                if self.debug:
                    logger.debug("%s: RPCServerError: %s", str(self), str(error))
                self._send_rpc_reply_error(error)
            except EOFError:
                if self.debug:
                    logger.debug("%s: Got EOF in reader_handle_message", str(self))
                error = ncerror.RPCSvrException(rpc, EOFError("EOF"))
                self._send_rpc_reply_error(error)
            except Exception as exception:
                if self.debug:
                    logger.debug("%s: Got unexpected exception in reader_handle_message: %s",
                                 str(self), str(exception))
                error = ncerror.RPCSvrException(rpc, exception)
                self._send_rpc_reply_error(error)


class NetconfMethods(object):
    """This is an abstract class that is used to document the server methods
    functionality.

    The base server code will return not-implemented if the method is not found
    in the methods object, so feel free to use duck-typing here (i.e., no need to
    inherit). Create a class that implements the rpc_* methods you handle and pass
    that to `NetconfSSHServer` init.
    """

    def nc_append_capabilities(self, capabilities):  # pylint: disable=W0613
        """This method should append any capabilities it supports to capabilities

        :param capabilities: The element to append capability elements to.
        :type capabilities: `lxml.Element`
        :return: None
        """
        return

    def rpc_get(self, session, rpc, filter_or_none):  # pylint: disable=W0613
        """Passed the filter element or None if not present

        :param session: The server session with the client.
        :type session: `NetconfServerSession`
        :param rpc: The topmost element in the received message.
        :type rpc: `lxml.Element`
        :param filter_or_none: The filter element if present.
        :type filter_or_none: `lxml.Element` or None
        :return: `lxml.Element` of "nc:data" type containing the requested state.
        :raises: `error.RPCServerError` which will be used to construct an XML error response.
        """
        raise ncerror.OperationNotSupportedProtoError(rpc)

    def rpc_get_config(self, session, rpc, source_elm, filter_or_none):  # pylint: disable=W0613
        """The client has requested the config state (config: true). The function is
        passed the source element and the filter element or None if not present

        :param session: The server session with the client.
        :type session: `NetconfServerSession`
        :param rpc: The topmost element in the received message.
        :type rpc: `lxml.Element`
        :param source_elm: The source element indicating where the config should be drawn from.
        :type source_elm: `lxml.Element`
        :param filter_or_none: The filter element if present.
        :type filter_or_none: `lxml.Element` or None
        :return: `lxml.Element` of "nc:data" type containing the requested state.
        :raises: `error.RPCServerError` which will be used to construct an XML error response.
        """
        raise ncerror.OperationNotSupportedProtoError(rpc)

    #---------------------------------------------------------------------------
    # These definitions will change to include required parameters like get and
    # get-config
    #---------------------------------------------------------------------------

    def rpc_copy_config(self, unused_session, rpc, *unused_params):
        """XXX API subject to change -- unfinished"""
        raise ncerror.OperationNotSupportedProtoError(rpc)

    def rpc_delete_config(self, unused_session, rpc, *unused_params):
        """XXX API subject to change -- unfinished"""
        raise ncerror.OperationNotSupportedProtoError(rpc)

    def rpc_edit_config(self, unused_session, rpc, *unused_params):
        """XXX API subject to change -- unfinished"""
        raise ncerror.OperationNotSupportedProtoError(rpc)

    def rpc_lock(self, unused_session, rpc, *unused_params):
        """XXX API subject to change -- unfinished"""
        raise ncerror.OperationNotSupportedProtoError(rpc)

    def rpc_unlock(self, unused_session, rpc, *unused_params):
        """XXX API subject to change -- unfinished"""
        raise ncerror.OperationNotSupportedProtoError(rpc)


class NetconfSSHServer(sshutil.server.SSHServer):
    """A netconf server.

    :param server_ctl: The object used for authenticating connections to the server.
    :type server_ctl: `ssh.ServerInterface`
    :param server_methods: An object which implements servers the rpc_* methods.
    :param port: The port to bind the server to.
    :param host_key: The file containing the host key.
    :param debug: True to enable debug logging.
    """

    def __init__(self, server_ctl=None, server_methods=None, port=830, host_key=None, debug=False):
        self.server_methods = server_methods if server_methods is not None else NetconfMethods()
        self.session_id = 1
        super(NetconfSSHServer, self).__init__(
            server_ctl,
            server_session_class=NetconfServerSession,
            port=port,
            host_key=host_key,
            debug=debug)

    def __del__(self):
        logger.error("Deleting %s", str(self))

    def _allocate_session_id(self):
        with self.lock:
            sid = self.session_id
            self.session_id += 1
            return sid

    def __str__(self):
        return "NetconfSSHServer(port={})".format(self.port)


__author__ = 'Christian Hopps'
__date__ = 'February 19 2015'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# January 12 2018, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2018, Deutsche Telekom AG.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes

from netconf.base import chunkit  # pylint: disable=W0611


def test_chunkit():
    """
    Handle all possible cases until we get firmly in the loop
    including coming out with leftovers

    >>> [x for x in chunkit("", 6, 3, "x")]                 # doctest: +ALLOW_UNICODE
    []
    >>> [x for x in chunkit("0", 6, 3, "x")]                # doctest: +ALLOW_UNICODE
    ['0xx']
    >>> [x for x in chunkit("01", 6, 3, "x")]               # doctest: +ALLOW_UNICODE
    ['01x']
    >>> [x for x in chunkit("012", 6, 3, "x")]              # doctest: +ALLOW_UNICODE
    ['012']
    >>> [x for x in chunkit("0123", 6, 3, "x")]             # doctest: +ALLOW_UNICODE
    ['0123']
    >>> [x for x in chunkit("01234", 6, 3, "x")]            # doctest: +ALLOW_UNICODE
    ['01234']
    >>> [x for x in chunkit("012345", 6, 3, "x")]           # doctest: +ALLOW_UNICODE
    ['012345']
    >>> [x for x in chunkit("0123456", 6, 3, "x")]          # doctest: +ALLOW_UNICODE
    ['0123', '456']
    >>> [x for x in chunkit("01234567", 6, 3, "x")]         # doctest: +ALLOW_UNICODE
    ['01234', '567']
    >>> [x for x in chunkit("012345678", 6, 3, "x")]        # doctest: +ALLOW_UNICODE
    ['012345', '678']
    >>> [x for x in chunkit("0123456789", 6, 3, "x")]       # doctest: +ALLOW_UNICODE
    ['012345', '6789']
    >>> [x for x in chunkit("0123456789a", 6, 3, "x")]      # doctest: +ALLOW_UNICODE
    ['012345', '6789a']
    >>> [x for x in chunkit("0123456789ab", 6, 3, "x")]     # doctest: +ALLOW_UNICODE
    ['012345', '6789ab']
    >>> [x for x in chunkit("0123456789abc", 6, 3, "x")]    # doctest: +ALLOW_UNICODE
    ['012345', '6789', 'abc']
    >>> [x for x in chunkit("0123456789abcd", 6, 3, "x")]   # doctest: +ALLOW_UNICODE
    ['012345', '6789a', 'bcd']
    >>> [x for x in chunkit("0123456789abcde", 6, 3, "x")]  # doctest: +ALLOW_UNICODE
    ['012345', '6789ab', 'cde']
    >>> [x for x in chunkit("0123456789abcdef", 6, 3, "x")]  # doctest: +ALLOW_UNICODE
    ['012345', '6789ab', 'cdef']
    >>> [x for x in chunkit("0123456789abcdefg", 6, 3, "x")]  # doctest: +ALLOW_UNICODE
    ['012345', '6789ab', 'cdefg']
    >>> [x for x in chunkit("0123456789abcdefgh", 6, 3, "x")]  # doctest: +ALLOW_UNICODE
    ['012345', '6789ab', 'cdefgh']
    >>> [x for x in chunkit("0123456789abcdefghi", 6, 3, "x")]  # doctest: +ALLOW_UNICODE
    ['012345', '6789ab', 'cdef', 'ghi']
    >>> [x for x in chunkit("0123456789abcdefghij", 6, 3, "x")]  # doctest: +ALLOW_UNICODE
    ['012345', '6789ab', 'cdefg', 'hij']
    >>> [x for x in chunkit("0123456789abcdefghijk", 6, 3, "x")]  # doctest: +ALLOW_UNICODE
    ['012345', '6789ab', 'cdefgh', 'ijk']
    >>> [x for x in chunkit("", 3)]                         # doctest: +ALLOW_UNICODE
    []
    >>> [x for x in chunkit("0", 3)]                        # doctest: +ALLOW_UNICODE
    ['0']
    >>> [x for x in chunkit("01", 3)]                       # doctest: +ALLOW_UNICODE
    ['01']
    >>> [x for x in chunkit("012", 3)]                      # doctest: +ALLOW_UNICODE
    ['012']
    >>> [x for x in chunkit("0123", 3)]                     # doctest: +ALLOW_UNICODE
    ['012', '3']
    >>> [x for x in chunkit("01234", 3)]                    # doctest: +ALLOW_UNICODE
    ['012', '34']
    >>> [x for x in chunkit("012345", 3)]                   # doctest: +ALLOW_UNICODE
    ['012', '345']
    >>> [x for x in chunkit("0123456", 3)]                  # doctest: +ALLOW_UNICODE
    ['012', '345', '6']
    >>> [x for x in chunkit("01234567", 3)]                 # doctest: +ALLOW_UNICODE
    ['012', '345', '67']
    >>> [x for x in chunkit("012345678", 3)]                # doctest: +ALLOW_UNICODE
    ['012', '345', '678']
    >>> [x for x in chunkit("0123456789", 3)]               # doctest: +ALLOW_UNICODE
    ['012', '345', '678', '9']
    >>> [x for x in chunkit("0123456789a", 3)]              # doctest: +ALLOW_UNICODE
    ['012', '345', '678', '9a']
    >>> [x for x in chunkit("0123456789ab", 3)]             # doctest: +ALLOW_UNICODE
    ['012', '345', '678', '9ab']
    >>> [x for x in chunkit("0123456789abc", 3)]            # doctest: +ALLOW_UNICODE
    ['012', '345', '678', '9ab', 'c']
    """
    """
    Don't handle this right now.
    >>> [x for x in chunkit("", 3, 3, "x")]
    []
    >>> [x for x in chunkit("0", 3, 3, "x")]
    ['0xx']
    >>> [x for x in chunkit("01", 3, 3, "x")]
    ['01x']
    >>> [x for x in chunkit("012", 3, 3, "x")]
    ['012']
    >>> [x for x in chunkit("0123", 3, 3, "x")]
    ['012', '3xx']
    >>> [x for x in chunkit("01234", 3, 3, "x")]
    ['012', '34x']
    >>> [x for x in chunkit("012345", 3, 3, "x")]
    ['012', '345']
    >>> [x for x in chunkit("0123456", 3, 3, "x")]
    ['012', '345', '6xx']
    >>> [x for x in chunkit("01234567", 3, 3, "x")]
    ['012', '345', '67x']
    >>> [x for x in chunkit("012345678", 3, 3, "x")]
    ['012', '345', '678']
    >>> [x for x in chunkit("0123456789", 3, 3, "x")]
    ['012', '345', '678', '9xx']
    >>> [x for x in chunkit("0123456789a", 3, 3, "x")]
    ['012', '345', '678', '9ax']
    >>> [x for x in chunkit("0123456789ab", 3, 3, "x")]
    ['012', '345', '678', '9ab']
    >>> [x for x in chunkit("0123456789abc", 3, 3, "x")]
    ['012', '345', '678', '9ab', 'cxx']
    """


__author__ = 'Christian Hopps'
__date__ = 'January 12 2018'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# February 19 2015, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import logging
import io
import socket
import sys
import threading
import traceback
from lxml import etree

from netconf import NSMAP, MAXSSHBUF
from netconf.error import ChannelClosed, FramingError, SessionError
import netconf.util as ncutil

logger = logging.getLogger(__name__)

NC_BASE_10 = "urn:ietf:params:netconf:base:1.0"
NC_BASE_11 = "urn:ietf:params:netconf:base:1.1"
XML_HEADER = """<?xml version="1.0" encoding="UTF-8"?>"""

if sys.version_info[0] >= 3:

    def lookahead(iterable):
        """Return an element and an indication if it's the last element"""
        i = iter(iterable)
        last = next(i)
        for e in i:
            yield last, False
            last = e
        yield last, True
else:

    def lookahead(iterable):
        """Return an element and an indication if it's the last element"""
        i = iter(iterable)
        last = i.next()
        for e in i:
            yield last, False
            last = e
        yield last, True


def chunkit(msg, maxsend, minsend=0, pad="\n"):
    """
    chunkit iterates over a msg returning chunks of at most maxsend
    size, and of at least minsend size if non-zero. Padding will be
    added if required. This function currently requires that maxsend
    is at least large enough to hold 2 minsend chunks.
    """
    # For now we'll make this assumption as it makes the
    # implementation much easier.
    assert maxsend >= 2 * minsend

    sz = len(msg)
    nchunks = sz // maxsend
    lastmax = sz % maxsend

    # Handle the special cases
    if sz == 0:
        return
    elif nchunks == 1 and lastmax == 0:
        yield msg
        return
    elif nchunks == 0:
        # lastmax == 0 then sz == 0 handled above.
        assert lastmax != 0
        if lastmax < minsend:
            msg = msg + pad * (minsend - lastmax)
        yield msg
        return

    # Make sure our final chunk is at least minsend long.
    nchunks -= 1
    penultmax = maxsend
    if lastmax == 0:
        lastmax = maxsend
        nchunks -= 1
    elif lastmax < minsend:
        penultmax -= minsend - lastmax
        lastmax = minsend

    left = 0
    for unused in range(0, nchunks):
        yield msg[left:left + maxsend]
        left += maxsend

    right = left + penultmax
    yield msg[left:right]
    yield msg[right:]


class NetconfTransportMixin(object):
    def connect(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


class NetconfPacketTransport(object):
    def send_pdu(self, msg, new_framing):
        raise NotImplementedError()

    def receive_pdu(self, new_framing):
        raise NotImplementedError()


class NetconfFramingTransport(NetconfPacketTransport):
    """Packetize an ssh stream into netconf PDUs -- doesn't need to be SSH specific"""

    def __init__(self, stream, max_chunk, debug):
        # XXX we have 2 channels defined one here and one in the connect/accept class
        self.stream = stream
        self.max_chunk = max_chunk
        self.debug = debug
        self.rbuffer = b""

    def __del__(self):
        self.close()

    def close(self):
        stream = self.stream
        if stream is not None:
            self.stream = None
            if self.debug:
                logger.debug("Closing netconf socket stream %s", str(stream))
            stream.close()

    def is_active(self):
        try:
            self.stream.is_active
        except AttributeError:
            transport = self.stream.get_transport()
            if not transport:
                return False
            return transport.is_active()
        else:
            return self.stream.is_active()

    def receive_pdu(self, new_framing):
        assert self.stream is not None
        if new_framing:
            return self._receive_11()
        else:
            return self._receive_10()

    def send_pdu(self, msg, new_framing):
        assert self.stream is not None
        if new_framing:
            bmsg = msg.encode('utf-8')
            blen = len(bmsg)
            msg = "\n#{}\n".format(blen).encode('utf-8') + bmsg + "\n##\n".encode('utf-8')
        else:
            msg += "]]>]]>"

        # Apparently ssh has a bug that requires minimum of 64 bytes?
        for chunk in chunkit(msg, self.max_chunk, 64):
            self.stream.sendall(chunk)

    def _receive_10(self):
        searchfrom = 0
        while True:
            eomidx = self.rbuffer.find(b"]]>]]>", searchfrom)
            if eomidx != -1:
                break
            searchfrom = max(0, len(self.rbuffer) - 5)
            buf = self.stream.recv(self.max_chunk)
            self.rbuffer += buf

        msg = self.rbuffer[:eomidx]
        self.rbuffer = self.rbuffer[eomidx + 6:]
        return msg.decode('utf-8')

    def _receive_chunk(self):
        blen = len(self.rbuffer)
        while blen < 4:
            buf = self.stream.recv(self.max_chunk)
            self.rbuffer += buf
            blen = len(self.rbuffer)
            if self.stream is None:
                if self.debug:
                    logger.debug("Channel closed: stream is None")
                raise ChannelClosed(self)
            if not buf:
                if self.debug:
                    logger.debug("Channel closed: Zero bytes read")
                raise ChannelClosed(self)

        if self.rbuffer[:2] != b"\n#":
            raise FramingError(self.rbuffer)
        self.rbuffer = self.rbuffer[2:]

        # Get chunk length or termination indicator
        idx = -1
        searchfrom = 0
        while True:
            idx = self.rbuffer.find(b"\n", searchfrom)
            if 12 > idx > 0:
                break
            if idx > 12 or len(self.rbuffer) > 12:
                raise FramingError(self.rbuffer)
            searchfrom = len(self.rbuffer)
            self.rbuffer += self.stream.recv(self.max_chunk)

        # Check for last chunk.
        if self.rbuffer[0:2] == b"#\n":
            self.rbuffer = self.rbuffer[2:]
            return None

        lenstr = self.rbuffer[:idx]
        self.rbuffer = bytes(self.rbuffer[idx + 1:])

        try:
            chunklen = int(lenstr)
            if not (4294967295 >= chunklen > 0):
                raise FramingError("Unacceptable chunk length: {}".format(chunklen))
        except ValueError:
            raise FramingError("Frame length not integer: {}".format(lenstr.encode('utf-8')))

        while True:
            blen = len(self.rbuffer)
            if blen >= chunklen:
                chunk = self.rbuffer[:chunklen]
                self.rbuffer = self.rbuffer[chunklen:]
                return chunk
            self.rbuffer += self.stream.recv(self.max_chunk)

    def _iter_receive_chunks(self):
        assert self.stream is not None
        chunk = self._receive_chunk()
        while chunk:
            yield chunk
            chunk = self._receive_chunk()

    def _receive_11(self):
        assert self.stream is not None
        data = b"".join([x for x in self._iter_receive_chunks()])
        return data.decode('utf-8')


class NetconfSession(object):
    """Netconf Protocol Server and Client"""

    # This class is almost idntical to sshutil.SSHServerSession We need to
    # figure a way to factor the commonality. One issue is that this class can
    # be used with any transport not just SSH so where should it go?

    def __init__(self, stream, debug, session_id, max_chunk=MAXSSHBUF):
        self.debug = debug
        self.pkt_stream = NetconfFramingTransport(stream, max_chunk, debug)
        self.new_framing = False
        self.capabilities = set()
        self.reader_thread = None
        self.lock = threading.Lock()
        self.session_id = session_id
        self.session_open = False

    def __del__(self):
        if hasattr(self, "session_open") and self.session_open:
            self.close()

    def is_active(self):
        with self.lock:
            return self.pkt_stream and self.pkt_stream.is_active()

    def __str__(self):
        return "NetconfSession(sid:{})".format(self.session_id)

    def send_message(self, msg):
        with self.lock:
            pkt_stream = self.pkt_stream
        if not pkt_stream:
            logger.info("Dropping message b/c no connection stream (%d): %s", len(msg), msg)
            return
        if self.debug:
            logger.debug("Sending message (%d): %s", len(msg), msg)
        pkt_stream.send_pdu(XML_HEADER + msg, self.new_framing)

    def _receive_message(self):
        # private method to receive a full message.
        with self.lock:
            if self.reader_thread and not self.reader_thread.keep_running:
                return None
            pkt_stream = self.pkt_stream
        return pkt_stream.receive_pdu(self.new_framing)

    def send_hello(self, caplist, session_id=None):
        msg = ncutil.elm("hello", attrib={'xmlns': NSMAP['nc']})
        caps = ncutil.elm("capabilities")
        for cap in caplist:
            ncutil.subelm(caps, "capability").text = str(cap)
        if session_id is not None:
            assert hasattr(self, "methods")
            self.methods.nc_append_capabilities(caps)  # pylint: disable=E1101
        msg.append(caps)

        if self.debug:
            logger.debug("%s: Sending HELLO", str(self))
        if session_id is not None:
            msg.append(ncutil.leaf_elm("session-id", str(session_id)))
        msg = etree.tostring(msg)
        self.send_message(msg.decode('utf-8'))

    def close(self):
        if self.debug:
            logger.debug("%s: Closing.", str(self))

        with self.lock:
            if self.session_open:
                self.session_open = False
                self.session_id = None

            if self.reader_thread:
                self.reader_thread.keep_running = False

            if self.pkt_stream is not None:
                if self.debug:
                    logger.debug("%s: Closing transport.", str(self))

                pkt_stream = self.pkt_stream
                self.pkt_stream = None

                if pkt_stream:
                    # If we are blocked on reading this should unblock us
                    pkt_stream.close()

    def _open_session(self, is_server):
        assert is_server or self.session_id is None

        # The transport should be connected at this point.
        try:
            # Send hello message.
            self.send_hello((NC_BASE_10, NC_BASE_11), self.session_id)

            # Get reply
            reply = self._receive_message()
            if self.debug:
                logger.debug("Received HELLO")

            # Parse reply
            tree = etree.parse(io.BytesIO(reply.encode('utf-8')))
            root = tree.getroot()
            caps = root.xpath("//nc:hello/nc:capabilities/nc:capability", namespaces=NSMAP)

            # Store capabilities
            for cap in caps:
                self.capabilities.add(cap.text.strip())

            if NC_BASE_11 in self.capabilities:
                self.new_framing = True
            elif NC_BASE_10 not in self.capabilities:
                who = "Server" if is_server else "Client"
                raise SessionError("{} doesn't implement 1.0 or 1.1 of netconf".format(who))

            # Get session ID.
            try:
                session_id = root.xpath("//nc:hello/nc:session-id", namespaces=NSMAP)[0].text
                # If we are a server it is a failure to receive a session id.
                if is_server:
                    raise SessionError("Client sent a session-id")
                self.session_id = int(session_id)
            except (KeyError, IndexError, AttributeError):
                if not is_server:
                    raise SessionError("Server didn't supply session-id")
            except ValueError:
                raise SessionError("Server supplied non integer session-id: {}".format(session_id))

            self.session_open = True

            # Create reader thread.
            self.reader_thread = threading.Thread(target=self._read_message_thread)
            self.reader_thread.daemon = True
            self.reader_thread.keep_running = True
            self.reader_thread.start()

            if self.debug:
                logger.debug("%s: Opened version %s session.", str(self), "1.1"
                             if self.new_framing else "1.0")

        except Exception:
            self.close()
            raise

    def _reader_exits(self):
        """This function is called from the session reader thread as it exits. No more
        messages will be read from the session socket.
        """
        raise NotImplementedError("reader_exits")

    def _reader_handle_message(self, msg):
        """This function is called from the session reader thread to process a received
        framed netconf message.
        """
        # Called from reader thread after receiving a framed message
        raise NotImplementedError("read_handle_message")

    def _read_message_thread(self):
        # XXX the locking and dealing with the exit of this thread needs improvement
        if self.debug:
            logger.debug("Starting reader thread.")

        reader_thread = self.reader_thread
        try:
            while self.pkt_stream:
                with self.lock:
                    pkt_stream = self.pkt_stream
                    if not reader_thread.keep_running:
                        break
                    assert pkt_stream is not None

                msg = self._receive_message()
                if msg:
                    self._reader_handle_message(msg)
                    closed = False
                else:
                    # Client closed, never really see this 1/2 open case unfortunately.
                    if self.debug:
                        logger.debug("Client remote closed, exiting reader thread.")
                    closed = True

                with self.lock:
                    if closed:
                        reader_thread.keep_running = False
                    if not reader_thread.keep_running:
                        break

            if self.debug:
                logger.debug("Exiting reader thread")

        except AttributeError as error:
            # Should we close the session cleanly or just disconnect?
            if "'NoneType' object has no attribute 'recv'" in str(error):
                logger.error("%s: Session channel cleared (open: %s): %s: %s", str(self),
                             str(self.session_open), str(error), traceback.format_exc())
            else:
                logger.error(
                    "Unexpected exception in reader thread [disconnecting+exiting]: %s: %s",
                    str(error), traceback.format_exc())
            self.close()
        except ChannelClosed as error:
            # Should we close the session cleanly or just disconnect?
            # if self.debug:
            #     logger.debug("%s: Session channel closed [session_open == %s]: %s: %s",
            #                  str(self),
            #                  str(self.session_open),
            #                  str(error),
            #                  traceback.format_exc())
            # else:
            logger.debug("%s: Session channel closed [session_open == %s]: %s", str(self),
                         str(self.session_open), str(error))
            try:
                self.close()
            except Exception as error:
                logger.debug("%s: Exception while closing during ChannelClosed: %s", str(self),
                             str(error))
        except SessionError as error:
            # Should we close the session cleanly or just disconnect?
            logger.error("%s Session error [closing session]: %s", str(self), str(error))
            self.close()
        except socket.error as error:
            if self.debug:
                logger.debug("Socket error in reader thread [exiting]: %s", str(error))
            self.close()
        except Exception as error:
            with self.lock:
                keep_running = reader_thread.keep_running
            if keep_running:
                logger.error(
                    "Unexpected exception in reader thread [disconnecting+exiting]: %s: %s",
                    str(error), traceback.format_exc())
                self.close()
            else:
                # XXX might want to catch errors due to disconnect and not re-raise
                logger.debug("Exception in reader thread [exiting]: %s: %s", str(error),
                             traceback.format_exc())
        finally:
            # If we are exiting the read thread we close the session.
            self._reader_exits()


__author__ = 'Christian Hopps'
__date__ = 'December 23 2014'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# February 17 2015, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import getpass
import logging
try:
    from lxml import etree
except ImportError:
    from xml.etree import ElementTree as etree
import paramiko as ssh

from sshutil import DisableGlobalCaching
from sshutil.cache import SSHConnectionCache, SSHNoConnectionCache
from netconf import client
from netconf import server
from netconf.error import RPCError, SessionError

logger = logging.getLogger(__name__)
nc_server = None
NC_PORT = None
SERVER_DEBUG = True
CLIENT_DEBUG = True


class NetconfMethods(server.NetconfMethods):
    def rpc_get(self, session, rpc, filter_or_none):
        del session  # unused
        del filter_or_none  # unused
        return etree.Element("ok")

    def rpc_get_config(self, session, rpc, source_elm, filter_or_none):
        del session  # unused
        del source_elm  # unused
        del filter_or_none  # unused
        return etree.Element("ok")

    def rpc_namespaced(self, unused_session, rpc, *unused_params):
        return etree.Element("ok")


def setup_module(unused_module):
    global nc_server

    logging.basicConfig(level=logging.DEBUG)
    DisableGlobalCaching()

    if nc_server is not None:
        logger.error("XXX Called setup_module multiple times")
    else:
        logger.info("Set up netconf server")
        server_ctl = server.SSHUserPassController(username=getpass.getuser(), password="admin")
        nc_server = server.NetconfSSHServer(
            server_ctl=server_ctl,
            server_methods=NetconfMethods(),
            port=NC_PORT,
            host_key="tests/host_key",
            debug=SERVER_DEBUG)


def cleanup_module(unused_module):
    global nc_server
    if nc_server:
        logger.info("Deleting server")

        # Delete the server so that we don't end up with a bunch of logging going on on exit.
        nc_server = None

        # now let's force garbage collection to try and get rid of other objects.
        logger.info("Garbage collecting")
        import gc
        gc.collect()


def test_bad_password():
    try:
        cache = SSHNoConnectionCache("SSH uncached connections")
        session = client.NetconfSSHSession(
            "127.0.0.1", password="badpass", port=nc_server.port, cache=cache)
    except ssh.AuthenticationException:
        pass
    else:
        logger.error("Unexpected success: %s", str(session))
        assert False


def test_not_supported():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = "<foobar/>"
    try:
        rval = session.send_rpc(query)
    except RPCError as error:
        assert error.get_error_tag() == "operation-not-supported"
    else:
        logger.error("Unexpected success: %s", str(rval))
        assert False


def test_namespaced_rpc():
    """TEST: Checked that namespaced RPCs work."""
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = '<namespaced xmlns="some:namespace:1.0"></namespaced>'
    rval = session.send_rpc(query)
    rval = session.send_rpc(query)
    assert rval
    # logger.debug("Get: {}", rval)
    session.close()


def test_malformed():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = "<get><foobar/></get><get/>"
    try:
        rval = session.send_rpc(query)
    except RPCError as error:
        assert error.get_error_tag() == "malformed-message"
    else:
        logger.error("Unexpected success: %s", str(rval))
        assert False


# XXX this hangs the server!
def test_malformed_2():
    session = client.NetconfSSHSession(
        "127.0.0.1", password="admin", port=nc_server.port, debug=CLIENT_DEBUG)
    assert session

    query = "</foobar>"
    try:
        rval = session.send_rpc(query)
    except RPCError as error:
        assert error.get_error_tag() == "malformed-message"
        session.close()
    except SessionError as error:
        # If the session closes that's OK too.
        pass
    else:
        logger.error("Unexpected success: %s", str(rval))
        assert False


def test_get():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = "<get><filter><status/></filter></get>"
    rval = session.send_rpc(query)
    assert rval
    # logger.debug("Get: {}", rval)
    session.close()


def test_get_config():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = "<get-config><source><running/></source></get-config>"
    rval = session.send_rpc(query)
    assert rval
    # logger.debug("Get: {}", rval)
    session.close()


def test_get_config_with_filter():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = "<get-config><source><running/></source><filter><foobar/></filter></get-config>"
    rval = session.send_rpc(query)
    assert rval
    session.close()


def test_get_config_missing_source():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = "<get-config></get-config>"
    try:
        rval = session.send_rpc(query)
    except RPCError as error:
        assert error.get_error_tag() == "missing-element"
    else:
        logger.error("Unexpected success: %s", str(rval))
        assert False
    session.close()


def test_get_config_with_non_filter():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session

    query = """<get-config><source><running/></source><foobar/></get-config>"""
    try:
        rval = session.send_rpc(query)
    except RPCError as error:
        assert error.get_error_tag() == "unknown-element"
    else:
        logger.error("Unexpected success: %s", str(rval))
        assert False
    session.close()


def test_close():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port)
    assert session
    session.close()


def test_multi_session():
    sessions = []
    for unused in range(0, 10):
        sessions.append(
            client.NetconfSSHSession("127.0.0.1", password="admin", port=nc_server.port))


def test_server_close():
    server_ctl = server.SSHUserPassController(username=getpass.getuser(), password="admin")
    for i in range(0, 10):
        logger.debug("Starting %d iteration", i)
        ns = server.NetconfSSHServer(
            server_ctl=server_ctl,
            server_methods=NetconfMethods(),
            port=None,
            host_key="tests/host_key",
            debug=SERVER_DEBUG)
        port = ns.port

        logger.info("Connect to server on port %d", port)
        session = client.NetconfSSHSession(
            "127.0.0.1", password="admin", port=port, debug=CLIENT_DEBUG)
        session.close()
        # NetconfSSHSession.flush()

        logger.debug("Closing")
        ns.close()
        logger.debug("Joining")
        ns.join()
    logger.debug("Test Complete")


def _test_multi_open(client_cache):

    logger.info("Create Server")
    server_ctl = server.SSHUserPassController(username=getpass.getuser(), password="admin")
    ns = server.NetconfSSHServer(
        server_ctl=server_ctl,
        server_methods=NetconfMethods(),
        port=NC_PORT,
        host_key="tests/host_key",
        debug=SERVER_DEBUG)
    port = ns.port

    logger.info("Open sessions")
    sessions = [
        client.NetconfSSHSession(
            "127.0.0.1", password="admin", port=port, debug=CLIENT_DEBUG, cache=client_cache)
        for unused in range(0, 25)
    ]

    logger.info("Close sessions")
    for session in sessions:
        session.close()

    logger.info("Reopening")
    sessions = [
        client.NetconfSSHSession(
            "127.0.0.1", password="admin", port=port, debug=CLIENT_DEBUG, cache=client_cache)
        for unused in range(0, 25)
    ]

    logger.info("Closeing")
    for session in sessions:
        session.close()

    logger.info("Reopening")
    sessions = [
        client.NetconfSSHSession(
            "127.0.0.1", password="admin", port=port, debug=CLIENT_DEBUG, cache=client_cache)
        for unused in range(0, 25)
    ]
    logger.info("Reclosing")
    for session in sessions:
        session.close()

    # Close down the server and join it to make sure it's closed
    logger.info("Closing server")
    ns.close()
    logger.info("Joining server")
    ns.join()

    # Delete the server so that we don't end up with a bunch of logging going on on exit.
    del ns
    del server_ctl


def test_multi_open_no_cache():
    _test_multi_open(SSHNoConnectionCache("SSH uncached connections"))


def test_multi_open_cache():
    _test_multi_open(SSHConnectionCache("test multi open cache", max_channels=50))


__author__ = 'Christian Hopps'
__date__ = 'February 17 2015'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# February 19 2015, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
from lxml import etree
from netconf import NSMAP


class NetconfException(Exception):
    pass


class ChannelClosed(NetconfException):
    pass


class FramingError(NetconfException):
    pass


class SessionError(NetconfException):
    pass


class ReplyTimeoutError(NetconfException):
    pass


class NetconfError(NetconfException):
    def __init__(self, output, tree, error):
        super(RPCError, self).__init__(output)
        self.tree = tree
        self.error = error

    def _get_error_val(self, value):
        try:
            return self.error.xpath("nc:" + value, namespaces=NSMAP)[0].text
        except IndexError:
            return None

    def get_error_tag(self):
        return self._get_error_val("error-tag")

    def get_error_type(self):
        return self._get_error_val("error-type")

    def get_error_info(self):
        return self._get_error_val("error-info")

    def get_error_severity(self):
        return self._get_error_val("error-severity")


# Backward compatible
RPCError = NetconfError

# RFC6241

# error-type
RPCERR_TYPE_TRANSPORT = 0
RPCERR_TYPE_RPC = 1
RPCERR_TYPE_PROTOCOL = 2
RPCERR_TYPE_APPLICATION = 3
RPCERR_TYPE_ENUM = {
    RPCERR_TYPE_TRANSPORT: "transport",
    RPCERR_TYPE_RPC: "rpc",
    RPCERR_TYPE_PROTOCOL: "protocol",
    RPCERR_TYPE_APPLICATION: "application"
}

# error-tag
RPCERR_TAG_IN_USE = "in-use"
RPCERR_TAG_INVALID_VALUE = "invalid-value"
RPCERR_TAG_TOO_BIG = "too-big"
RPCERR_TAG_MISSING_ATTRIBUTE = "missing-attribute"
RPCERR_TAG_BAD_ATTRIBUTE = "bad-attribute"
RPCERR_TAG_UNKNOWN_ATTRIBUTE = "unknown-attribute"
RPCERR_TAG_MISSING_ELEMENT = "missing-element"
RPCERR_TAG_BAD_ELEMENT = "bad-element"
RPCERR_TAG_UNKNOWN_ELEMENT = "unknown-element"
RPCERR_TAG_UNKNOWN_NAMESPACE = "unknown-namespace"
RPCERR_TAG_ACCESS_DENIED = "access-denied"
RPCERR_TAG_LOCK_DENIED = "lock-denied"
RPCERR_TAG_RESOURCE_DENIED = "resource-denied"
RPCERR_TAG_ROLLBACK_FAILED = "rollback-failed"
RPCERR_TAG_DATA_EXISTS = "data-exists"
RPCERR_TAG_DATA_MISSING = "data-missing"
RPCERR_TAG_OPERATION_NOT_SUPPORTED = "operation-not-supported"
RPCERR_TAG_OPERATION_FAILED = "operation-failed"
RPCERR_TAG_MALFORMED_MESSAGE = "malformed-message"
RPCERR_TAG_PARTIAL_OPERATION = "partial-operation"

# ============================================
# Server Errors for returning value to client.
# ============================================


class RPCServerError(NetconfException):
    def __init__(self, origmsg, etype, tag, **kwargs):
        # Add attrib and nsmap from original message.
        self.reply = etree.Element("rpc-reply", attrib=origmsg.attrib, nsmap=origmsg.nsmap)

        rpcerr = etree.SubElement(self.reply, "rpc-error")

        # We require a type, tag, and severity assuming error for severity.
        if etype in RPCERR_TYPE_ENUM:
            etype = RPCERR_TYPE_ENUM[etype]
        etree.SubElement(rpcerr, "error-type").text = str(etype)

        etree.SubElement(rpcerr, "error-tag").text = tag

        if "severity" not in kwargs:
            etree.SubElement(rpcerr, "error-severity").text = "error"

        # Now convert any other arguments to xml
        for key, value in kwargs.items():
            # Allow info to be a dictionary we convert to sub-elements
            if key == "info" and hasattr(value, "items"):
                infoelm = etree.SubElement(rpcerr, "error-info")
                for ikey, ivalue in value.items():
                    ikey = ikey.replace('_', '-')
                    etree.SubElement(infoelm, "{}".format(ikey)).text = str(ivalue)
            else:
                key = key.replace('_', '-')
                etree.SubElement(rpcerr, "error-{}".format(key)).text = str(value)

        # This sort of sucks for humans
        super(RPCServerError, self).__init__(self.get_reply_msg())

    def get_reply_msg(self):
        return etree.tounicode(self.reply)


class RPCSvrException(RPCServerError):
    def __init__(self, origmsg, exception, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            RPCERR_TYPE_PROTOCOL,
            RPCERR_TAG_OPERATION_FAILED,
            info=str(exception),
            **kwargs)


# Need to deprecate this as we are overriding a built-in
TimeoutError = ReplyTimeoutError  # pylint: disable=W0622

# -------------------------------------------------------------
# Netconf Mandated Errors - Ordered alphabetically by base type
# -------------------------------------------------------------


class _AccessDeniedError(RPCServerError):
    def __init__(self, origmsg, etype, **kwargs):
        RPCServerError.__init__(self, origmsg, etype, RPCERR_TAG_ACCESS_DENIED, **kwargs)


class AccessDeniedAppError(_AccessDeniedError):
    def __init__(self, origmsg, **kwargs):
        _AccessDeniedError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, **kwargs)


class AccessDeniedProtoError(_AccessDeniedError):
    def __init__(self, origmsg, **kwargs):
        _AccessDeniedError.__init__(self, origmsg, RPCERR_TYPE_PROTOCOL, **kwargs)


class _BadAttributeError(RPCServerError):
    def __init__(self, origmsg, element, attribute, etype, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            etype,
            RPCERR_TAG_BAD_ATTRIBUTE,
            info={'bad-element': element.tag,
                  'bad-attibute': attribute},
            **kwargs)


class BadAttributeAppError(_BadAttributeError):
    def __init__(self, origmsg, element, attribute, **kwargs):
        _BadAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_APPLICATION,
                                    **kwargs)


class BadAttributeProtoError(_BadAttributeError):
    def __init__(self, origmsg, element, attribute, **kwargs):
        _BadAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_PROTOCOL,
                                    **kwargs)


class BadAttributeRPCError(_BadAttributeError):
    def __init__(self, origmsg, element, attribute, **kwargs):
        _BadAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_RPC, **kwargs)


class _BadElementError(RPCServerError):
    def __init__(self, origmsg, element, etype, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            etype,
            RPCERR_TAG_BAD_ELEMENT,
            info={'bad-element': element.tag},
            **kwargs)


class BadElementAppError(_BadElementError):
    def __init__(self, origmsg, element, **kwargs):
        _BadElementError.__init__(self, origmsg, element, RPCERR_TYPE_APPLICATION, **kwargs)


class BadElementProtoError(_BadElementError):
    def __init__(self, origmsg, element, **kwargs):
        _BadElementError.__init__(self, origmsg, element, RPCERR_TYPE_PROTOCOL, **kwargs)


class DataExistsAppError(RPCServerError):
    def __init__(self, origmsg, **kwargs):
        RPCServerError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, RPCERR_TAG_DATA_EXISTS,
                                **kwargs)


class DataMissingAppError(RPCServerError):
    def __init__(self, origmsg, **kwargs):
        RPCServerError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, RPCERR_TAG_DATA_MISSING,
                                **kwargs)


class _InvalidValueError(RPCServerError):
    def __init__(self, origmsg, etype, **kwargs):
        RPCServerError.__init__(self, origmsg, etype, RPCERR_TAG_INVALID_VALUE, **kwargs)


class InvalidValueAppError(_InvalidValueError):
    def __init__(self, origmsg, **kwargs):
        _InvalidValueError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, **kwargs)


class InvalidValueProtoError(_InvalidValueError):
    def __init__(self, origmsg, **kwargs):
        _InvalidValueError.__init__(self, origmsg, RPCERR_TYPE_PROTOCOL, **kwargs)


class LockDeniedProtoError(RPCServerError):
    def __init__(self, origmsg, session_id, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            RPCERR_TYPE_PROTOCOL,
            RPCERR_TAG_LOCK_DENIED,
            info={'session-id': str(session_id)},
            **kwargs)


class MalformedMessageRPCError(RPCServerError):
    """
    If the server raises this exception the and netconf 1.0 is in use,
    the session will be closed
    """

    def __init__(self, origmsg):
        RPCServerError.__init__(self, origmsg, RPCERR_TYPE_RPC, RPCERR_TAG_MALFORMED_MESSAGE)


class _MissingAttributeError(RPCServerError):
    def __init__(self, origmsg, element, attribute, etype, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            etype,
            RPCERR_TAG_MISSING_ATTRIBUTE,
            info={'bad-element': element.tag,
                  'bad-attibute': attribute},
            **kwargs)


class MissingAttributeAppError(_MissingAttributeError):
    def __init__(self, origmsg, element, attribute, **kwargs):
        _MissingAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_APPLICATION,
                                        **kwargs)


class MissingAttributeProtoError(_MissingAttributeError):
    def __init__(self, origmsg, element, attribute, **kwargs):
        _MissingAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_PROTOCOL,
                                        **kwargs)


class MissingAttributeRPCError(_MissingAttributeError):
    def __init__(self, origmsg, element, attribute, **kwargs):
        _MissingAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_RPC,
                                        **kwargs)


class _MissingElementError(RPCServerError):
    def __init__(self, origmsg, tag, etype, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            etype,
            RPCERR_TAG_MISSING_ELEMENT,
            info={'bad-element': str(tag)},
            **kwargs)


class MissingElementAppError(_MissingElementError):
    def __init__(self, origmsg, tag, **kwargs):
        _MissingElementError.__init__(self, origmsg, tag, RPCERR_TYPE_APPLICATION, **kwargs)


class MissingElementProtoError(_MissingElementError):
    def __init__(self, origmsg, tag, **kwargs):
        _MissingElementError.__init__(self, origmsg, tag, RPCERR_TYPE_PROTOCOL, **kwargs)


class _OperationFailedError(RPCServerError):
    def __init__(self, origmsg, etype, **kwargs):
        RPCServerError.__init__(self, origmsg, etype, RPCERR_TAG_OPERATION_FAILED, **kwargs)


class OperationFailedAppError(_OperationFailedError):
    def __init__(self, origmsg, etype, **kwargs):
        _OperationFailedError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, **kwargs)


class OperationFailedProtoError(_OperationFailedError):
    def __init__(self, origmsg, **kwargs):
        _OperationFailedError.__init__(self, origmsg, RPCERR_TYPE_PROTOCOL, **kwargs)


class OperationFailedRPCError(_OperationFailedError):
    def __init__(self, origmsg, **kwargs):
        _OperationFailedError.__init__(self, origmsg, RPCERR_TYPE_RPC, **kwargs)


class _OperationNotSupportedError(RPCServerError):
    def __init__(self, origmsg, etype, **kwargs):
        RPCServerError.__init__(self, origmsg, etype, RPCERR_TAG_OPERATION_NOT_SUPPORTED, **kwargs)


class OperationNotSupportedAppError(_OperationNotSupportedError):
    def __init__(self, origmsg, **kwargs):
        _OperationNotSupportedError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, **kwargs)


class OperationNotSupportedProtoError(_OperationNotSupportedError):
    def __init__(self, origmsg, **kwargs):
        _OperationNotSupportedError.__init__(self, origmsg, RPCERR_TYPE_PROTOCOL, **kwargs)


# # This is a pretty complex error if a server really supports it it can implement it itself.
# class _PartialOperationError(RPCServerError):
#     def __init__(self, origmsg, ok_elms, err_elms, noop_elms, **kwargs):
#         RPCServerError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION,
#                                 RPCERR_TAG_PARTIAL_OPERATION,
#                                 **kwargs)


class _ResourceDeniedError(RPCServerError):
    def __init__(self, origmsg, etype, **kwargs):
        RPCServerError.__init__(self, origmsg, etype, RPCERR_TAG_RESOURCE_DENIED, **kwargs)


class ResourceDeniedAppError(_ResourceDeniedError):
    def __init__(self, origmsg, **kwargs):
        _ResourceDeniedError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, **kwargs)


class ResourceDeniedProtoError(_ResourceDeniedError):
    def __init__(self, origmsg, **kwargs):
        _ResourceDeniedError.__init__(self, origmsg, RPCERR_TYPE_PROTOCOL, **kwargs)


class ResourceDeniedRPCError(_ResourceDeniedError):
    def __init__(self, origmsg, **kwargs):
        _ResourceDeniedError.__init__(self, origmsg, RPCERR_TYPE_RPC, **kwargs)


class ResourceDeniedTransportError(_ResourceDeniedError):
    def __init__(self, origmsg, **kwargs):
        _ResourceDeniedError.__init__(self, origmsg, RPCERR_TYPE_TRANSPORT, **kwargs)


class _RollbackFailedError(RPCServerError):
    def __init__(self, origmsg, etype, **kwargs):
        RPCServerError.__init__(self, origmsg, etype, RPCERR_TAG_ROLLBACK_FAILED, **kwargs)


class RollbackFailedAppError(_RollbackFailedError):
    def __init__(self, origmsg, **kwargs):
        _RollbackFailedError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, **kwargs)


class RollbackFailedProtoError(_RollbackFailedError):
    def __init__(self, origmsg, **kwargs):
        _RollbackFailedError.__init__(self, origmsg, RPCERR_TYPE_PROTOCOL, **kwargs)


class _TooBigError(RPCServerError):
    def __init__(self, origmsg, etype, **kwargs):
        RPCServerError.__init__(self, origmsg, etype, RPCERR_TAG_TOO_BIG, **kwargs)


class TooBigAppError(_TooBigError):
    def __init__(self, origmsg, **kwargs):
        _TooBigError.__init__(self, origmsg, RPCERR_TYPE_APPLICATION, **kwargs)


class TooBigProtoError(_TooBigError):
    def __init__(self, origmsg, **kwargs):
        _TooBigError.__init__(self, origmsg, RPCERR_TYPE_PROTOCOL, **kwargs)


class TooBigRPCError(_TooBigError):
    def __init__(self, origmsg, **kwargs):
        _TooBigError.__init__(self, origmsg, RPCERR_TYPE_RPC, **kwargs)


class TooBigTransportError(_TooBigError):
    def __init__(self, origmsg, **kwargs):
        _TooBigError.__init__(self, origmsg, RPCERR_TYPE_TRANSPORT, **kwargs)


class _UnknownAttributeError(RPCServerError):
    def __init__(self, origmsg, element, attribute, etype, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            etype,
            RPCERR_TAG_UNKNOWN_ATTRIBUTE,
            info={'bad-element': element.tag,
                  'bad-attibute': attribute},
            **kwargs)


class UnknownAttributeAppError(_UnknownAttributeError):
    def __init__(self, origmsg, element, attribute, etype, **kwargs):
        _UnknownAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_APPLICATION,
                                        **kwargs)


class UnknownAttributeProtoError(_UnknownAttributeError):
    def __init__(self, origmsg, element, attribute, etype, **kwargs):
        _UnknownAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_PROTOCOL,
                                        **kwargs)


class UnknownAttributeRPCError(_UnknownAttributeError):
    def __init__(self, origmsg, element, attribute, etype, **kwargs):
        _UnknownAttributeError.__init__(self, origmsg, element, attribute, RPCERR_TYPE_RPC,
                                        **kwargs)


class _UnknownElementError(RPCServerError):
    def __init__(self, origmsg, element, etype, **kwargs):
        RPCServerError.__init__(
            self,
            origmsg,
            etype,
            RPCERR_TAG_UNKNOWN_ELEMENT,
            info={'bad-element': element.tag},
            **kwargs)


class UnknownElementAppError(_UnknownElementError):
    def __init__(self, origmsg, element, **kwargs):
        _UnknownElementError.__init__(self, origmsg, element, RPCERR_TYPE_APPLICATION, **kwargs)


class UnknownElementProtoError(_UnknownElementError):
    def __init__(self, origmsg, element, **kwargs):
        _UnknownElementError.__init__(self, origmsg, element, RPCERR_TYPE_PROTOCOL, **kwargs)


class _UnknownNamespaceError(RPCServerError):
    def __init__(self, origmsg, element, etype, **kwargs):
        try:
            qname = etree.QName(element.tag)
            tag = qname.localname
            ns = qname.namespace
        except Exception:
            tag = element.tag
            ns = "no namespace map"
        RPCServerError.__init__(
            self,
            origmsg,
            etype,
            RPCERR_TAG_UNKNOWN_NAMESPACE,
            info={'bad-element': tag,
                  'bad-namespace': ns},
            **kwargs)


class UnknownNamespaceAppError(_UnknownNamespaceError):
    def __init__(self, origmsg, element, etype, **kwargs):
        _UnknownNamespaceError.__init__(self, origmsg, element, RPCERR_TYPE_APPLICATION, **kwargs)


class UnknownNamespaceProtoError(_UnknownNamespaceError):
    def __init__(self, origmsg, element, etype, **kwargs):
        _UnknownNamespaceError.__init__(self, origmsg, element, RPCERR_TYPE_PROTOCOL, **kwargs)


# Backward compat -- XXX need some deprecation warning for these.
RPCSvrBadElement = BadElementAppError
RPCSvrErrNotImpl = OperationNotSupportedProtoError
RPCSvrErrBadMsg = MalformedMessageRPCError
RPCSvrInvalidValue = InvalidValueProtoError
RPCSvrMissingElement = MissingElementAppError
RPCSvrUnknownElement = UnknownElementAppError

__author__ = 'Christian Hopps'
__date__ = 'February 19 2015'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# December 23 2014, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
from lxml.etree import register_namespace

MAXSSHBUF = 16 * 1024
NSMAP = {}


def nsmap_add(prefix, namespace):
    "Add a prefix namespace mapping to the modules mapping dictionary"
    NSMAP[prefix] = namespace
    register_namespace(prefix, namespace)


def nsmap_update(nsdict):
    "Add a dicitonary of prefx namespace mappings to the modules mapping dictionary"
    NSMAP.update(nsdict)
    for key, val in nsdict.items():
        register_namespace(key, val)


def qmap(key):
    return "{" + NSMAP[key] + "}"


# Add base spec namespace
nsmap_add('nc', "urn:ietf:params:xml:ns:netconf:base:1.0")



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# December 23 2014, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import logging
import netconf.client as client
from netconf.error import NetconfError
from mockserver import init_mock_server

logger = logging.getLogger(__name__)
NC_PORT = None
NC_DEBUG = True


def setup_module(unused_module):
    global NC_PORT
    NC_PORT = init_mock_server()


def test_query():
    query = """
    <get>
    <filter type="subtree">
    <devices xmlns="http://tail-f.com/ns/ncs">
    <global-settings/>
    </devices>
    </filter>
    </get>
    """

    query = """
    <get>
    <filter type="xpath" select="/devices/global-settings"/>
    </get>
    """
    logger.info("Connecting to 127.0.0.1 port %d", NC_PORT)
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=NC_PORT, debug=NC_DEBUG)
    session.send_rpc(query)


def test_bad_query():
    session = client.NetconfSSHSession("127.0.0.1", password="admin", port=NC_PORT, debug=NC_DEBUG)
    try:
        unused, unused, output = session.send_rpc("<get><unknown/></get>")
        logger.warning("Got unexpected output: %s", str(output))
    except NetconfError:
        pass


def test_context_manager():
    select = """
    <devices xmlns="http://tail-f.com/ns/ncs">
    <global-settings/>
    </devices>
    """
    select = "/devices/global-settings"
    logger.info("Connecting to 127.0.0.1 port %d", NC_PORT)
    with client.connect_ssh("127.0.0.1", password="admin", port=NC_PORT, debug=NC_DEBUG) as session:
        session.get(select)


__author__ = 'Christian Hopps'
__date__ = 'December 23 2014'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# January 22 2018, Christian E. Hopps <chopps@gmail.com>
#
# Copyright (c) 2018, Deutsche Telekom AG.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
import getpass
import logging
import re
from lxml import etree
import netconf.client
from mockserver import init_mock_server

logger = logging.getLogger(__name__)
NC_PORT = None
NC_DEBUG = False


def setup_module(unused_module):
    global NC_PORT
    NC_PORT = init_mock_server()


def test_get_config():
    session = netconf.client.NetconfSSHSession(
        "127.0.0.1", username=getpass.getuser(), password="admin", port=NC_PORT, debug=NC_DEBUG)
    result = session.get_config()
    result = etree.tostring(result, encoding="unicode")
    result = re.sub(">[ \t\n\r]*<", "><", result).strip()
    expected = """<data xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"><interfaces><interface><name>Ethernet0/0</name><shutdown>true</shutdown></interface><interface><name>Ethernet0/1</name><shutdown>false</shutdown></interface><interface><name>FastEthernet1/0</name><shutdown>false</shutdown></interface><interface><name>FastEthernet1/1</name><shutdown>false</shutdown><name>GigabitEthernet2/0</name><shutdown>false</shutdown></interface></interfaces></data>"""
    assert result == expected


def test_get():
    session = netconf.client.NetconfSSHSession(
        "127.0.0.1", username=getpass.getuser(), password="admin", port=NC_PORT, debug=NC_DEBUG)
    result = session.get()
    result = etree.tostring(result, encoding="unicode")
    result = re.sub(">[ \t\n\r]*<", "><", result).strip()
    expected = """<data xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"><interfaces><interface><name>Ethernet0/0</name><shutdown>true</shutdown><state>down</state></interface><interface><name>Ethernet0/1</name><shutdown>false</shutdown><state>down</state></interface><interface><name>FastEthernet1/0</name><shutdown>false</shutdown><state>up</state></interface><interface><name>FastEthernet1/1</name><shutdown>false</shutdown><state>down</state></interface></interfaces></data>"""
    assert result == expected


__author__ = 'Christian E. Hopps'
__date__ = 'January 22 2018'
__version__ = '1.0'
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 eval: (yapf-mode 1) -*-
#
# February 19 2015, Christian Hopps <chopps@gmail.com>
#
# Copyright (c) 2015, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division, unicode_literals, print_function, nested_scopes
from contextlib import contextmanager
import logging
import io
import threading
import socket

import sshutil.conn
from lxml import etree
from monotonic import monotonic
from netconf import NSMAP
from netconf.base import NetconfSession
from netconf.error import RPCError, SessionError, ReplyTimeoutError
from netconf import util

logger = logging.getLogger(__name__)


def _is_filter(select):
    return select.lstrip().startswith("<")


def _get_selection(elm, select):
    if not select:
        return

    # XXX Do we want to add any namespace stuff to the filter node here?

    felm = util.elm("nc:filter")
    if hasattr(select, "nsmap"):
        felm.attrib["type"] = "subtree"
        felm.append(select)
    elif _is_filter(select):
        felm.attrib["type"] = "subtree"
        felm.append(etree.fromstring(select))
    else:
        felm.attrib["type"] = "xpath"
        felm.attrib["select"] = select

    elm.append(felm)


class Timeout(object):
    def __init__(self, timeout):
        self.start_time = monotonic()
        if timeout is None:
            self.end_time = None
        else:
            self.end_time = self.start_time + timeout

    def is_expired(self):
        if self.end_time is None:
            return False
        return self.end_time < monotonic()

    def remaining(self):
        if self.end_time is None:
            return None
        ctime = monotonic()
        if self.end_time < ctime:
            return 0
        else:
            return self.end_time - ctime


class NetconfClientSession(NetconfSession):
    """Netconf Protocol"""

    def __init__(self, stream, debug=False):
        super(NetconfClientSession, self).__init__(stream, debug, None)
        self.message_id = 0
        self.closing = False
        self.rpc_out = {}

        # Condition to handle rpc_out queue
        self.cv = threading.Condition()

        super(NetconfClientSession, self)._open_session(False)

    def __str__(self):
        return "NetconfClientSession(sid:{})".format(self.session_id)

    def close(self):
        """Close the session."""

        if self.debug:
            logger.debug("%s: Closing session.", str(self))

        reply = None
        try:
            # So we need a lock here to check these members.
            send = False
            with self.cv:
                if self.session_id is not None and self.is_active():
                    send = True

            if send:
                self.send_rpc_async("<close-session/>", noreply=True)
                # Don't wait for a reply the session is closed!
        except socket.error:
            if self.debug:
                logger.debug("Got socket error sending close-session request, ignoring")

        super(NetconfClientSession, self).close()

        if self.debug:
            logger.debug("%s: Closed: %s", str(self), str(reply))

    def is_reply_ready(self, msg_id):
        """Check whether reply is ready (or session closed)"""
        with self.cv:
            if not self.is_active():
                raise SessionError("Session closed while checking for reply")
            return self.rpc_out[msg_id] is not None

    def wait_reply(self, msg_id, timeout=None):
        """Wait for a reply to a given RPC message ID.

        :param msg_id: the RPC message ID returned from one of the async method calls
        :return: (Message as an lxml tree, Parsed reply content, Parsed message content).
        :rtype: (lxml.etree, lxml.Element, lxml.Element)
        :raises: RPCError, SessionError
        """
        assert msg_id in self.rpc_out

        check_timeout = Timeout(timeout)

        self.cv.acquire()
        # XXX need to make sure the channel doesn't close
        while self.rpc_out[msg_id] is None and self.is_active():
            remaining = check_timeout.remaining()

            self.cv.wait(remaining)
            if self.rpc_out[msg_id] is not None:
                break

            if check_timeout.is_expired():
                raise ReplyTimeoutError(
                    "Timeout ({}s) while waiting for RPC reply to msg-id: {}".format(
                        timeout, msg_id))

        if not self.is_active():
            self.cv.release()
            raise SessionError("Session closed while waiting for reply")

        tree, reply, msg = self.rpc_out[msg_id]
        del self.rpc_out[msg_id]
        self.cv.release()

        error = reply.xpath("nc:rpc-error", namespaces=NSMAP)
        if error:
            raise RPCError(msg, tree, error[0])

        # data = reply.xpath("nc:data", namespaces=self.nsmap)
        # ok = reply.xpath("nc:ok", namespaces=self.nsmap)
        return tree, reply, msg

    def send_rpc_async(self, rpc, noreply=False):
        """Send a generic RPC to the server and await the reply.

        :param rpc: The XML of the netconf RPC, not including the <rpc> tag.
        :type rpc: str or `lxml.Element`
        :param noreply: True if no reply is required.
        :type noreply: Boolean

        :return: The RPC message id which can be passed to wait_reply for the results.
        """
        # Not sure it makes sense to go back to a string here, but OK.
        # Need to be a bit careful about namespaces the default needs to be nc:
        if hasattr(rpc, "nsmap"):
            rpc = etree.tounicode(rpc)

        # Get the next message id
        with self.cv:
            assert self.session_id is not None
            msg_id = self.message_id
            self.message_id += 1

        if self.debug:
            logger.debug("%s: Sending RPC message-id: %s", str(self), str(msg_id))

        def sendit():
            self.send_message("""<rpc message-id="{}"
                xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">{}</rpc>""".format(msg_id, rpc))

        if noreply:
            sendit()
            return None

        with self.cv:
            sendit()
            # Mark us as expecting a reply
            self.rpc_out[msg_id] = None

        return msg_id

    def send_rpc(self, rpc, timeout=None):
        """Send a generic RPC to the server and await the reply.

        :param rpc (string): The XML of the netconf RPC, not including the <rpc> tag.
        :return: (Message as an lxml tree, Parsed reply content, Parsed message content).
        :rtype: (lxml.etree, lxml.Element, lxml.Element)
        :raises: RPCError, SessionError
        """
        msg_id = self.send_rpc_async(rpc)
        return self.wait_reply(msg_id, timeout)

    def edit_config_async(self, target, mode, newconf):
        """Operate on ~config~ in ~target~ using ~newconf~ according to ~mode~ ("merge", "replace" or
        "none"). If "none" then no nodes are modified until a element specifies the mode as an
        attribute.

        :param target: the target of the config, defaults to "running".
        :param method: "merge", "replace" or "none"
        :param newconf: The new configuration.
        :return: The RPC message id which can be passed to wait_reply for the results.
        :raises: SessionError
        """
        if hasattr(target, "nsmap"):
            target = target.tag

        rpc = """
<edit-config>
  <target>
    <""" + target + """/>
  </target>
"""
        if mode is not None and mode != "":
            rpc += "  <default-operation>{}</default-operation>\n".format(mode)
        rpc += newconf
        rpc += "</edit-config>\n"
        return self.send_rpc_async(rpc)

    def edit_config(self, target="running", method="", newconf="", timeout=None):
        """Operate on ~config~ in ~target~ using ~newconf~ according to ~mode~ ("merge", "replace" or
        "none"). If "none" then no nodes are modified until a element specifies the mode as an
        attribute.

        :param target: the target of the config, defaults to "running".
        :param method: "merge" (netconf default), "replace" or "none".
        :param newconf: The new configuration.
        :param timeout: A value in fractional seconds to wait for the operation to complete or
                        `None` for no timeout.
        :return: The result of the edit operation
        :rtype: lxml.Element
        :raises: ReplyTimeoutError, RPCError, SessionError
        """
        msg_id = self.edit_config_async(target, method, newconf)
        _, reply, _ = self.wait_reply(msg_id, timeout)
        return reply

    def get_config_async(self, source, select):
        """Get config asynchronously for a given source from the server. If `select` is
        specified it is either an XPATH expression or XML subtree filter for
        selecting a subsection of the config.

        :param source: the source of the config, defaults to "running".
        :param select: An XML subtree filter or XPATH expression to select a subsection of config.
        :return: The RPC message id which can be passed to wait_reply for the results.
        :raises: SessionError
        """
        getelm = util.elm("get-config")
        if not hasattr(source, "nsmap"):
            source = util.elm(source)
        util.subelm(util.subelm(getelm, "source"), source)
        _get_selection(getelm, select)
        return self.send_rpc_async(getelm)

    def get_config(self, source="running", select=None, timeout=None):
        """Get config for a given source from the server. If `select` is specified it
        is either an XPATH expression or XML subtree filter for selecting a
        subsection of the config. If `timeout` is not `None` it specifies how
        long to wait for the get operation to complete.

        :param source: the source of the config, defaults to "running".
        :param select: An XML subtree filter or XPATH expression to select a subsection of config.
        :param timeout: A value in fractional seconds to wait for the operation to complete or
                        `None` for no timeout.
        :return: The Parsed XML config (i.e., "<config>...</config>".)
        :rtype: lxml.Element
        :raises: ReplyTimeoutError, RPCError, SessionError
        """
        msg_id = self.get_config_async(source, select)
        _, reply, _ = self.wait_reply(msg_id, timeout)
        return reply.find("nc:data", namespaces=NSMAP)

    def get_async(self, select):
        """Get operational state asynchronously from the server. If `select` is
        specified it is either an XPATH expression or XML subtree filter for
        selecting a subsection of the state. If `timeout` is not `None` it
        specifies how long to wait for the get operation to complete.

        :param select: A XML subtree filter or XPATH expression to select a subsection of state.
        :return: The RPC message id which can be passed to wait_reply for the results.
        :raises: SessionError
        """

        getelm = util.elm("get")
        _get_selection(getelm, select)
        return self.send_rpc_async(getelm)

    def get(self, select=None, timeout=None):
        """Get operational state from the server. If `select` is specified it is either
        an XPATH expression or XML subtree filter for selecting a subsection of
        the state. If `timeout` is not `None` it specifies how long to wait for
        the get operation to complete.

        :param select: A XML subtree filter or XPATH expression to select a subsection of state.
        :param timeout: A value in fractional seconds to wait for the operation to complete or
                       `None` for no timeout.
        :return: The Parsed XML state (i.e., "<data>...</data>".)
        :rtype: lxml.Element
        :raises: ReplyTimeoutError, RPCError, SessionError
        """
        msg_id = self.get_async(select)
        _, reply, _ = self.wait_reply(msg_id, timeout)
        return reply.find("nc:data", namespaces=NSMAP)

    # ----------------
    # Internal Methods
    # ----------------

    def _reader_exits(self):
        """This function is called from the session reader thread as it exits. No more
        messages will be read from the session socket.
        """
        if self.debug:
            logger.debug("%s: Reader thread exited notifying all.", str(self))
        with self.cv:
            self.cv.notify_all()

    def _reader_handle_message(self, msg):
        """This function is called from the session reader thread to process a received
        framed netconf message.
        """
        try:
            tree = etree.parse(io.BytesIO(msg.encode('utf-8')))
            if not tree:
                raise SessionError(msg, "Invalid XML from server.")
        except etree.XMLSyntaxError:
            raise SessionError(msg, "Invalid XML from server.")

        replies = tree.xpath("/nc:rpc-reply", namespaces=NSMAP)
        if not replies:
            raise SessionError(msg, "No rpc-reply found")

        for reply in replies:
            try:
                msg_id = int(reply.get('message-id'))
            except (TypeError, ValueError):
                # # Cisco is returning errors without message-id attribute which
                # # is non-rfc-conforming it is doing this for any malformed XML
                # # not simply missing message-id attribute.
                # error = reply.xpath("nc:rpc-error", namespaces=self.nsmap)
                # if error:
                #     raise RPCError(received, tree, error[0])
                raise SessionError(msg, "No valid message-id attribute found")

            # Queue the message
            with self.cv:
                try:
                    if msg_id not in self.rpc_out:
                        if self.debug:
                            logger.debug("Ignoring unwanted reply for message-id %s", str(msg_id))
                        return
                    elif self.rpc_out[msg_id] is not None:
                        logger.warning("Received multiple replies for message-id %s:"
                                       " before: %s now: %s", str(msg_id), str(
                                           self.rpc_out[msg_id]), str(msg))

                    if self.debug:
                        logger.debug("%s: Received rpc-reply message-id: %s", str(self),
                                     str(msg_id))
                    self.rpc_out[msg_id] = tree, reply, msg
                except Exception as error:
                    logger.debug("%s: Unexpected exception: %s", str(self), str(error))
                    raise
                finally:
                    self.cv.notify_all()


class NetconfSSHSession(NetconfClientSession):
    def __init__(self,
                 host,
                 port=830,
                 username=None,
                 password=None,
                 debug=False,
                 cache=None,
                 proxycmd=None):
        """A netconf SSH client session.

        If `username` is not specified then it will be obtained with
        getpass.getuser(). If an ssh agent is available it will be used for
        authentication. A users .ssh/config will be processed for making the ssh
        connection and any proxycmd found therein will also be utilized.

        :param host: The host to connect to.
        :param port: The port to connect to.
        :param username: The username to connect with. If not specified getpass.getuser()
                         will be used.
        :param password: The password or passkey to authenticate with.
        :param debug: Enable debug logging
        :param cache: An SSH cache (`sshutil.cache`) to use for caching connections.
        :param proxycmd: A proxy command string for connecting with
        """
        if username is None:
            import getpass
            username = getpass.getuser()
        stream = sshutil.conn.SSHClientSession(
            host, port, "netconf", username, password, debug, cache=cache, proxycmd=proxycmd)
        super(NetconfSSHSession, self).__init__(stream, debug)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


@contextmanager
def connect_ssh(host,
                port=830,
                username=None,
                password=None,
                debug=False,
                cache=None,
                proxycmd=None):
    """A context manager method for opening a netconf SSH session.

    If `username` is not specified then it will be obtained with
    getpass.getuser(). If an ssh agent is available it will be used for
    authentication. A users .ssh/config will be processed for making the ssh
    connection and any proxycmd found therein will also be utilized.

    :param host: The host to connect to.
    :param port: The port to connect to.
    :param username: The username to connect with. If not specified getpass.getuser() will be used
    :param password: The password or passkey to authenticate with.
    :param debug: Enable debug logging
    :param cache: An SSH cache (`sshutil.cache`) to use for caching connections.
    :param proxycmd: A proxy command string for connecting with
    """
    session = NetconfSSHSession(host, port, username, password, debug, cache, proxycmd)
    yield session
    session.close()


__author__ = 'Christian Hopps'
__date__ = 'February 19 2015'
__version__ = '1.0'
__docformat__ = "restructuredtext en"
